import ccp
import streamlit as st
import json
import pandas as pd
import io
import base64
import zipfile
import toml
import time
import sentry_sdk
import logging
from ccp.compressor import Point1Sec, StraightThrough
from ccp.config.utilities import r_getattr
from ccp.config.units import ureg
from pathlib import Path

# import everything that is common to ccp_app_straight_through and ccp_app_back_to_back
from ccp.app.common import (
    pressure_units,
    specific_heat_units,
    density_units,
    polytropic_methods,
    parameters_map,
    oil_iso_options,
    specific_heat_calculate,
    density_calculate,
    get_gas_composition,
    to_excel,
    convert,
)

sentry_sdk.init(
    dsn="https://8fd0e79dffa94dbb9747bf64e7e55047@o348313.ingest.sentry.io/4505046640623616",
    # Set traces_sample_rate to 1.0 to capture 100%
    # of transactions for performance monitoring.
    # We recommend adjusting this value in production.
    traces_sample_rate=1.0,
    auto_enabling_integrations=False,
)


def main():
    """The code has to be inside this main function to allow sentry to work."""
    Q_ = ccp.Q_

    assets = Path(__file__).parent / "assets"
    ccp_ico = assets / "favicon.ico"
    ccp_logo = assets / "ccp.png"
    css_path = assets / "style.css"
    with open(css_path, "r") as f:
        css = f.read()

    st.set_page_config(
        page_title="ccp",
        page_icon=str(ccp_ico),
        layout="wide",
    )

    def image_base64(im):
        with open(im, "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read()).decode()
        encoded_image = "data:image/png;base64," + encoded_string
        html_string = f'<img src="data:image/png;base64,{encoded_string}" style="text-align: center" width="250">'
        return html_string

    with st.sidebar.container():
        st.sidebar.markdown(image_base64(ccp_logo), unsafe_allow_html=True)

    st.markdown(f"<style>{css}</style>", unsafe_allow_html=True)

    title_alignment = """
    <p style="text-align: center; font-weight: bold; font-size:20px;">
     ccp 
    </p>
    """
    st.sidebar.markdown(title_alignment, unsafe_allow_html=True)
    st.markdown(
        """
    ## Performance Test Straight-Through Compressor
    """
    )

    def get_session():
        # if file is being loaded, session_name is set with the file name
        if "session_name" not in st.session_state:
            st.session_state.session_name = ""
        if "straight_through" not in st.session_state:
            st.session_state.straight_through = ""
        if "expander_state" not in st.session_state:
            st.session_state.expander_state = False
        if "ccp_version" not in st.session_state:
            st.session_state.ccp_version = ccp.__version__
        if "app_type" not in st.session_state:
            st.session_state.app_type = "straight_through"
        for curve in ["head", "power", "eff", "discharge_pressure"]:
            if f"fig_{curve}" not in st.session_state:
                st.session_state[f"fig_{curve}"] = ""
        if "bearing_mechanical_losses" not in st.session_state:
            st.session_state.bearing_mechanical_losses = False
        if "oil_specific_heat" not in st.session_state:
            st.session_state.oil_specific_heat = False
        if "oil_iso" not in st.session_state:
            st.session_state.oil_iso = False

    get_session()

    # Create a Streamlit sidebar with a file uploader to load a session state file
    with st.sidebar.expander("📁 File"):
        with st.form("my_form", clear_on_submit=False):
            st.session_state.session_name = st.text_input(
                "Session name",
                value=st.session_state.session_name,
            )

            file = st.file_uploader("📂 Open File", type=["ccp"])
            submitted = st.form_submit_button("Load")
            save_button = st.form_submit_button("💾 Save")

        if submitted and file is not None:
            st.write("Loaded!")
            # open file with zip

            with zipfile.ZipFile(file) as my_zip:
                # get the ccp version
                try:
                    version = my_zip.read("ccp.version").decode("utf-8")
                except KeyError:
                    version = "0.3.5"

                for name in my_zip.namelist():
                    if name.endswith(".json"):
                        session_state_data = convert(
                            json.loads(my_zip.read(name)), version
                        )
                        if "flow_point_guarantee" not in session_state_data:
                            raise ValueError("File is not a ccp straight-through file.")

                # extract figures and straight_through objects
                for name in my_zip.namelist():
                    if name.endswith(".png"):
                        session_state_data[name.split(".")[0]] = my_zip.read(name)
                    elif name.endswith(".toml"):
                        # create file object to read the toml file
                        straight_through_file = convert(
                            io.StringIO(my_zip.read(name).decode("utf-8")), version
                        )
                        session_state_data[name.split(".")[0]] = StraightThrough.load(
                            straight_through_file
                        )

            session_state_data_copy = session_state_data.copy()
            # remove keys that cannot be set with st.session_state.update
            for key in session_state_data.keys():
                if key.startswith(
                    ("FormSubmitter", "my_form", "uploaded", "form", "table")
                ):
                    del session_state_data_copy[key]
            st.session_state.update(session_state_data_copy)
            st.session_state.session_name = file.name.replace(".ccp", "")
            st.session_state.expander_state = True
            st.rerun()

        if save_button:
            session_state_dict = dict(st.session_state)

            # create a zip file to add the data to
            file_name = f"{st.session_state.session_name}.ccp"
            session_state_dict_copy = session_state_dict.copy()
            with zipfile.ZipFile(file_name, "w") as my_zip:
                my_zip.writestr("ccp.version", ccp.__version__)
                # first save figures
                for key, value in session_state_dict.items():
                    if isinstance(
                        value, (bytes, st.runtime.uploaded_file_manager.UploadedFile)
                    ):
                        if key.startswith("fig"):
                            my_zip.writestr(f"{key}.png", value)
                        del session_state_dict_copy[key]
                    if isinstance(value, StraightThrough):
                        my_zip.writestr(
                            f"{key}.toml", toml.dumps(value._dict_to_save())
                        )
                        del session_state_dict_copy[key]
                # then save the rest of the session state
                session_state_json = json.dumps(session_state_dict_copy)
                my_zip.writestr("session_state.json", session_state_json)

            with open(file_name, "rb") as file:
                st.download_button(
                    label="💾 Save As",
                    data=file,
                    file_name=file_name,
                    mime="application/json",
                )

    # Gas selection
    fluid_list = []
    for fluid in ccp.fluid_list.keys():
        fluid_list.append(fluid.lower())
        for possible_name in ccp.fluid_list[fluid].possible_names:
            if possible_name != fluid.lower():
                fluid_list.append(possible_name)
    fluid_list = sorted(fluid_list)
    fluid_list.insert(0, "")

    with st.form(key="form_gas_selection", enter_to_submit=False, border=False):
        with st.expander("Gas Selection", expanded=st.session_state.expander_state):
            gas_compositions_table = {}
            gas_columns = st.columns(6)
            for i, gas_column in enumerate(gas_columns):
                gas_compositions_table[f"gas_{i}"] = {}

                gas_compositions_table[f"gas_{i}"]["name"] = gas_column.text_input(
                    "Gas Name",
                    value=f"gas_{i}",
                    key=f"gas_{i}",
                    help=(
                        """
                    Gas name will be selected in Data Sheet and Test Data.

                    Fill in gas components and molar fractions for each gas.
                    """
                        if i == 0
                        else None
                    ),
                )

                default_components = [
                    "methane",
                    "ethane",
                    "propane",
                    "n-butane",
                    "i-butane",
                    "n-pentane",
                    "i-pentane",
                    "n-hexane",
                    "n-heptane",
                    "n-octane",
                    "n-nonane",
                    "nitrogen",
                    "h2s",
                    "co2",
                    "h2o",
                ]

                gas_composition_list = []
                for key in st.session_state:
                    if "compositions_table" in key:
                        for column in st.session_state[key][f"gas_{i}"]:
                            if "component" in column:
                                idx = column.split("_")[1]
                                gas_composition_list.append(
                                    {
                                        "component": st.session_state[key][f"gas_{i}"][
                                            column
                                        ],
                                        "molar_fraction": st.session_state[key][
                                            f"gas_{i}"
                                        ][f"molar_fraction_{idx}"],
                                    }
                                )
                if not gas_composition_list:
                    gas_composition_list = [
                        {"component": molecule, "molar_fraction": 0.0}
                        for molecule in default_components
                    ]

                gas_composition_df = pd.DataFrame(gas_composition_list)
                gas_composition_df_edited = gas_column.data_editor(
                    gas_composition_df,
                    num_rows="dynamic",
                    key=f"table_gas_{i}_composition",
                    height=int((len(default_components) + 1) * 37.35),
                    use_container_width=True,
                    column_config={
                        "component": st.column_config.SelectboxColumn(
                            st.session_state[f"gas_{i}"],
                            options=fluid_list,
                            width="small",
                        ),
                        "molar_fraction": st.column_config.NumberColumn(
                            "mol %",
                            min_value=0.0,
                            default=0.0,
                            required=True,
                            format="%.3f",
                        ),
                    },
                )

                for column in gas_composition_df_edited:
                    for j, value in enumerate(gas_composition_df_edited[column]):
                        gas_compositions_table[f"gas_{i}"][f"{column}_{j}"] = value

            submit_composition = st.form_submit_button("Submit", type="primary")

            if "gas_compositions_table" not in st.session_state or submit_composition:
                st.session_state["gas_compositions_table"] = gas_compositions_table

    # add container with 4 columns and 2 rows
    with st.sidebar.expander("⚙️ Options"):
        reynolds_correction = st.checkbox("Reynolds Correction", value=True)
        casing_heat_loss = st.checkbox("Casing Heat Loss", value=True)
        bearing_mechanical_losses = st.checkbox("Bearing Mechanical Losses", value=True)
        calculate_leakages = st.checkbox("Calculate Leakages", value=True)
        seal_gas_flow = st.checkbox("Seal Gas Flow", value=True)

        # add a disabled flag in the parameters_map dict based on the checkbox
        if seal_gas_flow:
            parameters_map["seal_gas_flow_m"]["disabled"] = False
            parameters_map["seal_gas_temperature"]["disabled"] = False
        else:
            parameters_map["seal_gas_flow_m"]["disabled"] = True
            parameters_map["seal_gas_temperature"]["disabled"] = True
            parameters_map["seal_gas_flow_m"]["value"] = ""
            parameters_map["seal_gas_temperature"]["value"] = ""

        variable_speed = st.checkbox("Variable Speed", value=True)
        show_points = st.checkbox(
            "Show Points",
            value=True,
            help="If marked, show points in the plotted curves in addition to interpolation.",
        )
        # add text input for the ambient pressure
        st.text("Ambient Pressure")
        ambient_pressure_magnitude_col, ambient_pressure_unit_col = st.columns(2)
        with ambient_pressure_magnitude_col:
            ambient_pressure_magnitude = st.text_input(
                "Ambient Pressure",
                value=1.01325,
                key="ambient_pressure_magnitude",
                label_visibility="collapsed",
            )
        with ambient_pressure_unit_col:
            ambient_pressure_unit = st.selectbox(
                "Unit",
                options=pressure_units,
                index=pressure_units.index("bar"),
                key="ambient_pressure_unit",
                label_visibility="collapsed",
            )
        ambient_pressure = Q_(
            float(ambient_pressure_magnitude), ambient_pressure_unit
        ).to("bar")
        ureg.define(f"barg = 1 * bar; offset: {ambient_pressure.magnitude}")

        def on_oil_specific_heat_change():
            if st.session_state.oil_specific_heat:
                st.session_state.oil_iso = False

        def on_oil_iso_change():
            if st.session_state.oil_iso:
                st.session_state.oil_specific_heat = False

        # add text input for the test lube oil specific heat
        st.text("Test Lube Oil")
        # select box for oil specific heat input
        (
            oil_specific_heat_checkbox_col,
            oil_specific_heat_magnitude_col,
            oil_specific_heat_unit_col,
        ) = st.columns([0.5, 0.2, 0.3])
        with oil_specific_heat_checkbox_col:
            oil_specific_heat = st.checkbox(
                "Specific Heat",
                key="oil_specific_heat",
                value=False,
                on_change=on_oil_specific_heat_change,
                help="If marked, uses this oil specific heat "
                "and density for bearing mechanical losses calculation "
                "and disables ISO oil classification.",
            )
        with oil_specific_heat_magnitude_col:
            oil_specific_heat_magnitude = st.text_input(
                "Oil Specific Heat",
                value=2.03,
                key="oil_specific_heat_magnitude",
                label_visibility="collapsed",
                disabled=not st.session_state.oil_specific_heat,
            )
        with oil_specific_heat_unit_col:
            oil_specific_heat_unit = st.selectbox(
                "Unit",
                options=specific_heat_units,
                index=specific_heat_units.index("kJ/kg/degK"),
                key="oil_specific_heat_unit",
                label_visibility="collapsed",
                disabled=not st.session_state.oil_specific_heat,
            )

        oil_density_col, oil_density_magnitude_col, oil_density_unit_col = st.columns(
            [0.5, 0.2, 0.3]
        )
        with oil_density_col:
            st.text("Density")
        with oil_density_magnitude_col:
            oil_density_magnitude = st.text_input(
                "Oil Density",
                value=846.9,
                key="oil_density_magnitude",
                label_visibility="collapsed",
                disabled=not st.session_state.oil_specific_heat,
            )
        with oil_density_unit_col:
            oil_density_unit = st.selectbox(
                "Unit",
                options=density_units,
                index=density_units.index("kg/m³"),
                key="oil_density_unit",
                label_visibility="collapsed",
                disabled=not st.session_state.oil_specific_heat,
            )
        if oil_specific_heat:
            oil_specific_heat_value = Q_(
                float(oil_specific_heat_magnitude), oil_specific_heat_unit
            ).to("kJ/kg/kelvin")
            oil_density_value = Q_(float(oil_density_magnitude), oil_density_unit).to(
                "kg/m³"
            )

        # select box for oil ISO classification input
        oil_iso_checkbox_col, oil_iso_select_box_col = st.columns([0.6, 0.4])
        with oil_iso_checkbox_col:
            oil_iso = st.checkbox(
                "Oil ISO Classification",
                key="oil_iso",
                on_change=on_oil_iso_change,
                help="If marked, uses the ISO oil classification "
                "for bearing mechanical losses calculation "
                "and disables specific heat and density input.",
            )
        with oil_iso_select_box_col:
            oil_iso_classification = st.selectbox(
                "ISO",
                options=oil_iso_options,
                index=oil_iso_options.index("VG 32"),
                key="oil_iso_classification",
                label_visibility="collapsed",
                disabled=not st.session_state.oil_iso,
            )

        st.text("Polytropic method")
        polytropic_method = st.selectbox(
            "Polytropic",
            options=polytropic_methods.keys(),
            index=list(polytropic_methods.keys()).index("Sandberg-Colby"),
            key="polytropic_method",
            label_visibility="collapsed",
        )

    ccp.config.POLYTROPIC_METHOD = polytropic_methods[polytropic_method]

    # add dict to store the values for guarantee and test points
    # in the parameters_map
    number_of_test_points = 6
    points_list = ["point_guarantee"] + [
        f"point_{i}" for i in range(1, number_of_test_points + 1)
    ]
    points_fo_list = [f"point_fo_{i}" for i in range(1, number_of_test_points + 1)]

    for parameter in parameters_map:
        if not parameter.endswith("_fo"):
            parameters_map[parameter]["points"] = {
                "data_sheet_units": None,
                "test_units": None,
            }
            for point in points_list:
                parameters_map[parameter]["points"][point] = {
                    "value": None,
                }
        else:
            parameters_map[parameter]["points_fo"] = {
                "test_fo_units": None,
            }
            for point in points_fo_list:
                parameters_map[parameter]["points_fo"][point] = {
                    "value": None,
                }

    with st.expander("Data Sheet", expanded=st.session_state.expander_state):
        # build container with 8 columns
        points_title = st.container()
        points_title_columns = points_title.columns(3, gap="small")
        points_title_columns[0].markdown("")
        points_title_columns[1].markdown("Units")
        points_title_columns[2].markdown("")

        points_gas_columns = points_title.columns(3, gap="small")
        points_gas_columns[0].markdown("Gas Selection")
        points_gas_columns[1].markdown("")
        gas_options = [
            st.session_state[f"gas_{i}"] for i, gas in enumerate(gas_columns)
        ]

        # fill the gas selection dropdowns with the gases selected
        def get_index_selected_gas(gas_name):
            try:
                index_gas_name = gas_options.index(st.session_state[gas_name])
            except (KeyError, ValueError):
                index_gas_name = 0
            return index_gas_name

        gas_name_point_guarantee = points_gas_columns[2].selectbox(
            "gas_point_guarantee",
            options=gas_options,
            label_visibility="collapsed",
            key="gas_point_guarantee",
            index=get_index_selected_gas("gas_point_guarantee"),
        )

        # build one container with 8 columns for each parameter
        for parameter in [
            "flow",
            "suction_pressure",
            "suction_temperature",
            "discharge_pressure",
            "discharge_temperature",
            "power",
            "power_shaft",
            "speed",
            "head",
            "eff",
            "b",
            "D",
            "surface_roughness",
            "casing_area",
        ]:
            parameter_container = st.container()
            (
                parameters_col,
                units_col,
                values_col,
            ) = parameter_container.columns(3, gap="small")

            parameters_col.markdown(
                f"{parameters_map[parameter]['label']}",
                help=f"{parameters_map[parameter].get('help', '')}",
            )

            parameters_map[parameter]["points"]["data_sheet_units"] = (
                units_col.selectbox(
                    f"{parameter} units",
                    options=parameters_map[parameter]["units"],
                    key=f"{parameter}_units_point_guarantee",
                    label_visibility="collapsed",
                )
            )
            parameters_map[parameter]["points"]["point_guarantee"]["value"] = (
                values_col.text_input(
                    f"{parameter} value.",
                    key=f"{parameter}_point_guarantee",
                    label_visibility="collapsed",
                )
            )

    with st.expander("Curves", expanded=st.session_state.expander_state):
        # add upload button for curve
        # check if fig_dict was created when loading state. Otherwise, create it
        plot_limits = {}
        fig_dict_uploaded = {}

        for curve in ["head", "eff", "discharge_pressure", "power"]:
            st.markdown(f"### {parameters_map[curve]['label']}")
            parameter_container = st.container()
            curves_col = parameter_container

            plot_limits[curve] = {}
            # create upload button
            fig_dict_uploaded[f"uploaded_fig_{curve}"] = curves_col.file_uploader(
                f"Upload {curve}.",
                type=["png"],
                key=f"uploaded_fig_{curve}",
                label_visibility="collapsed",
            )

            if fig_dict_uploaded[f"uploaded_fig_{curve}"] is not None:
                st.session_state[f"fig_{curve}"] = fig_dict_uploaded[
                    f"uploaded_fig_{curve}"
                ].read()

            # add container to x range
            for axis in ["x", "y"]:
                with st.container():
                    (
                        plot_limit,
                        units_col,
                        lower_value_col,
                        upper_value_col,
                    ) = curves_col.columns(4, gap="small")
                    plot_limit.markdown(f"{axis} range")
                    plot_limits[curve][f"{axis}"] = {}
                    plot_limits[curve][f"{axis}"]["lower_limit"] = (
                        lower_value_col.text_input(
                            "Lower limit",
                            key=f"{axis}_{curve}_lower",
                            label_visibility="collapsed",
                        )
                    )
                    plot_limits[curve][f"{axis}"]["upper_limit"] = (
                        upper_value_col.text_input(
                            "Upper limit",
                            key=f"{axis}_{curve}_upper",
                            label_visibility="collapsed",
                        )
                    )
                    if axis == "x":
                        plot_limits[curve][f"{axis}"]["units"] = units_col.selectbox(
                            f"{parameter} units",
                            options=parameters_map["flow_v"]["units"],
                            key=f"{axis}_{curve}_flow_units",
                            label_visibility="collapsed",
                        )
                    else:
                        plot_limits[curve][f"{axis}"]["units"] = units_col.selectbox(
                            f"{parameter} units",
                            options=parameters_map[curve]["units"],
                            key=f"{axis}_{curve}_units",
                            label_visibility="collapsed",
                        )

    number_of_test_points = 6
    number_of_columns = number_of_test_points + 2

    with st.expander("Test Data", expanded=st.session_state.expander_state):
        # add title
        number_of_columns = number_of_test_points + 2
        # build container with 8 columns
        points_title = st.container()
        points_title_columns = points_title.columns(number_of_columns, gap="small")
        for i, col in enumerate(points_title_columns):
            if i == 0:
                col.markdown("")
            elif i == 1:
                col.markdown("Units")
            else:
                col.markdown(f"Point {i - 1}")

        # gas selection
        point_gas = st.container()
        point_gas_columns = point_gas.columns(number_of_columns, gap="small")
        for i, col in enumerate(point_gas_columns):
            if i == 0:
                col.markdown("Gas Selection")
            elif i == 1:
                col.markdown("")
            else:
                gas_options = [
                    st.session_state[f"gas_{i}"] for i, gas in enumerate(gas_columns)
                ]
                col.selectbox(
                    f"gas_point_{i - 1}",
                    options=gas_options,
                    label_visibility="collapsed",
                    key=f"gas_point_{i - 1}",
                    index=get_index_selected_gas(f"gas_point_{i - 1}"),
                )

        # build one container with 8 columns for each parameter
        for parameter in [
            "flow",
            "suction_pressure",
            "suction_temperature",
            "discharge_pressure",
            "discharge_temperature",
            "casing_delta_T",
            "speed",
            "balance_line_flow_m",
            "seal_gas_flow_m",
            "seal_gas_temperature",
            "oil_flow_journal_bearing_de",
            "oil_flow_journal_bearing_nde",
            "oil_flow_thrust_bearing_nde",
            "oil_inlet_temperature",
            "oil_outlet_temperature_de",
            "oil_outlet_temperature_nde",
        ]:
            parameter_container = st.container()
            parameter_columns = parameter_container.columns(
                number_of_columns, gap="small"
            )
            for i, col in enumerate(parameter_columns):
                if i == 0:
                    col.markdown(
                        f"{parameters_map[parameter]['label']}",
                        help=f"{parameters_map[parameter].get('help', '')}",
                    )
                elif i == 1:
                    parameters_map[parameter]["points"]["test_units"] = col.selectbox(
                        f"{parameter} units",
                        options=parameters_map[parameter]["units"],
                        key=f"{parameter}_units",
                        label_visibility="collapsed",
                        disabled=parameters_map[parameter].get("disabled", False),
                    )
                else:
                    parameters_map[parameter]["points"][f"point_{i - 1}"]["value"] = (
                        col.text_input(
                            f"{parameter} value.",
                            key=f"{parameter}_point_{i - 1}",
                            label_visibility="collapsed",
                            disabled=parameters_map[parameter].get("disabled", False),
                        )
                    )

    with st.expander("Flowrate Calculation", expanded=st.session_state.expander_state):
        # add title
        number_of_columns = number_of_test_points + 2
        # build container with 8 columns
        points_title = st.container()
        points_title_columns = points_title.columns(number_of_columns, gap="small")
        for i, col in enumerate(points_title_columns):
            if i == 0:
                col.markdown("")
            elif i == 1:
                col.markdown("Units")
            else:
                col.markdown(f"Point {i - 1}")

        # gas selection
        point_gas = st.container()
        point_gas_columns = point_gas.columns(number_of_columns, gap="small")
        for i, col in enumerate(point_gas_columns):
            if i == 0:
                col.markdown("Gas Selection")
            elif i == 1:
                col.markdown("")
            else:
                gas_options = [
                    st.session_state[f"gas_{i}"] for i, gas in enumerate(gas_columns)
                ]
                col.selectbox(
                    f"gas_fo_{i - 1}",
                    options=gas_options,
                    label_visibility="collapsed",
                    key=f"gas_fo_{i - 1}",
                    index=get_index_selected_gas(f"gas_fo_{i - 1}"),
                )

        # build one container with 8 columns for each parameter
        for parameter in [
            "outer_diameter_fo",
            "inner_diameter_fo",
            "upstream_pressure_fo",
            "upstream_temperature_fo",
            "pressure_drop_fo",
            "tappings_fo",
            "mass_flow_fo",
        ]:
            parameter_container = st.container()
            parameter_columns = parameter_container.columns(
                number_of_columns, gap="small"
            )
            for i, col in enumerate(parameter_columns):
                if i == 0:
                    col.markdown(
                        f"{parameters_map[parameter]['label']}",
                        help=f"{parameters_map[parameter].get('help', '')}",
                    )
                elif i == 1:
                    if parameter == "tappings_fo":
                        pass
                    else:
                        parameters_map[parameter]["points_fo"]["test_fo_units"] = (
                            col.selectbox(
                                f"{parameter} units",
                                options=parameters_map[parameter]["units"],
                                key=f"{parameter}_units",
                                label_visibility="collapsed",
                            )
                        )
                else:
                    if parameter == "tappings_fo":
                        parameters_map[parameter]["points_fo"]["test_fo_units"] = (
                            col.selectbox(
                                f"{parameter} value",
                                options=parameters_map[parameter]["units"],
                                key=f"{parameter}_{i - 1}",
                                label_visibility="collapsed",
                            )
                        )
                    else:
                        # elif parameter != "mass_flow_fo":
                        parameters_map[parameter]["points_fo"][f"point_fo_{i - 1}"][
                            "value"
                        ] = col.text_input(
                            f"{parameter} value.",
                            key=f"{parameter}_{i - 1}",
                            label_visibility="collapsed",
                        )

    # add calculate button
    straight_through = None

    def fo_calc():
        global missed_points
        kwargs = {}
        missed_points = []

        for i, col in enumerate(parameter_columns):
            if i > 1:
                if (
                    st.session_state[f"outer_diameter_fo_{i-1}"] == ""
                    or st.session_state[f"inner_diameter_fo_{i-1}"] == ""
                    or st.session_state[f"upstream_pressure_fo_{i-1}"] == ""
                    or st.session_state[f"upstream_temperature_fo_{i-1}"] == ""
                    or st.session_state[f"pressure_drop_fo_{i-1}"] == ""
                ):
                    missed_points.append(i - 1)
                    parameters_map["mass_flow_fo"]["points_fo"][f"point_fo_{i - 1}"][
                        "value"
                    ] = ""
                    st.session_state[f"mass_flow_fo_{i-1}"] = ""
                    continue
                else:
                    kwargs["state"] = ccp.State(
                        p=Q_(
                            float(st.session_state[f"upstream_pressure_fo_{i-1}"]),
                            parameters_map["upstream_pressure_fo"]["points_fo"][
                                "test_fo_units"
                            ],
                        ),
                        T=Q_(
                            float(st.session_state[f"upstream_temperature_fo_{i-1}"]),
                            parameters_map["upstream_temperature_fo"]["points_fo"][
                                "test_fo_units"
                            ],
                        ),
                        fluid=get_gas_composition(
                            st.session_state[f"gas_fo_{i-1}"],
                            gas_compositions_table,
                            default_components,
                        ),
                    )

                    kwargs["delta_p"] = Q_(
                        float(st.session_state[f"pressure_drop_fo_{i-1}"]),
                        parameters_map["pressure_drop_fo"]["points_fo"][
                            "test_fo_units"
                        ],
                    )

                    kwargs["D"] = Q_(
                        float(st.session_state[f"outer_diameter_fo_{i-1}"]),
                        parameters_map["outer_diameter_fo"]["points_fo"][
                            "test_fo_units"
                        ],
                    )

                    kwargs["d"] = Q_(
                        float(st.session_state[f"inner_diameter_fo_{i-1}"]),
                        parameters_map["inner_diameter_fo"]["points_fo"][
                            "test_fo_units"
                        ],
                    )

                    kwargs["tappings"] = st.session_state[f"tappings_fo_{i-1}"]

                    fo_temp = ccp.FlowOrifice(**kwargs)
                    parameters_map["mass_flow_fo"]["points_fo"][f"point_fo_{i - 1}"][
                        "value"
                    ] = fo_temp.qm.to(
                        parameters_map["mass_flow_fo"]["points_fo"]["test_fo_units"]
                    ).m
                    st.session_state[f"mass_flow_fo_{i-1}"] = str(
                        round(
                            parameters_map["mass_flow_fo"]["points_fo"][
                                f"point_fo_{i - 1}"
                            ]["value"],
                            5,
                        )
                    )
                st.session_state.missed_points = missed_points

    with st.container():
        calculate_col, calculate_speed_col, calculate_flowrate_col = st.columns(
            [1, 1, 1]
        )
        calculate_button = calculate_col.button(
            "Calculate",
            type="primary",
            use_container_width=True,
            help="Calculate results using the data sheet speed.",
        )
        calculate_speed_button = calculate_speed_col.button(
            "Calculate Speed",
            type="primary",
            use_container_width=True,
            help="Calculate speed to match the discharge pressure.",
        )
        calculate_flowrate = calculate_flowrate_col.button(
            "Calculate Flowrate",
            type="primary",
            use_container_width=True,
            on_click=fo_calc,
            help="Calculate flowrate with orifice plate data.",
        )

    if calculate_flowrate:
        calculate_flowrate = False
        progress_value = 0
        progress_bar = st.progress(progress_value, text="Calculating...")

        if "missed_points" not in st.session_state:
            st.session_state["missed_points"] = []

        for i in range(len(st.session_state["missed_points"])):
            st.warning(
                f"Missing data for point {st.session_state['missed_points'][i]}!"
            )
        time.sleep(0.1)
        progress_bar.progress(100, text="Done!")

    # create test points
    test_points = []
    kwargs = {}

    if calculate_button or calculate_speed_button:
        progress_value = 0
        progress_bar = st.progress(progress_value, text="Calculating...")
        # calculate guarantee point
        kwargs_guarantee = {}

        # get gas composition from selected gas
        gas_composition_data_sheet = {}
        gas_composition_data_sheet["point_guarantee"] = get_gas_composition(
            gas_name_point_guarantee,
            gas_compositions_table,
            default_components,
        )

        # to get data sheet units we use parameters_map[parameter]["selected_units"]
        if (
            Q_(0, parameters_map["flow"]["points"]["data_sheet_units"]).dimensionality
            == "[mass] / [time]"
        ):
            kwargs_guarantee["flow_m"] = Q_(
                float(st.session_state["flow_point_guarantee"]),
                parameters_map["flow"]["points"]["data_sheet_units"],
            )
        else:
            kwargs_guarantee["flow_v"] = Q_(
                float(st.session_state["flow_point_guarantee"]),
                parameters_map["flow"]["points"]["data_sheet_units"],
            )
        kwargs_guarantee["suc"] = ccp.State(
            p=Q_(
                float(st.session_state["suction_pressure_point_guarantee"]),
                parameters_map["suction_pressure"]["points"]["data_sheet_units"],
            ),
            T=Q_(
                float(st.session_state["suction_temperature_point_guarantee"]),
                parameters_map["suction_temperature"]["points"]["data_sheet_units"],
            ),
            fluid=gas_composition_data_sheet["point_guarantee"],
        )
        kwargs_guarantee["disch"] = ccp.State(
            p=Q_(
                float(st.session_state["discharge_pressure_point_guarantee"]),
                parameters_map["discharge_pressure"]["points"]["data_sheet_units"],
            ),
            T=Q_(
                float(
                    st.session_state["discharge_temperature_point_guarantee"],
                ),
                parameters_map["discharge_temperature"]["points"]["data_sheet_units"],
            ),
            fluid=gas_composition_data_sheet["point_guarantee"],
        )
        kwargs_guarantee["speed"] = Q_(
            float(st.session_state["speed_point_guarantee"]),
            parameters_map["speed"]["points"]["data_sheet_units"],
        )
        kwargs_guarantee["b"] = Q_(
            float(st.session_state["b_point_guarantee"]),
            parameters_map["b"]["points"]["data_sheet_units"],
        )
        kwargs_guarantee["D"] = Q_(
            float(st.session_state["D_point_guarantee"]),
            parameters_map["D"]["points"]["data_sheet_units"],
        )

        power_guarantee = Q_(
            float(st.session_state["power_point_guarantee"]),
            parameters_map["power"]["points"]["data_sheet_units"],
        )
        if bearing_mechanical_losses:
            if st.session_state["power_shaft_point_guarantee"] == "":
                power_shaft_guarantee = Q_(
                    float(st.session_state["power_point_guarantee"]),
                    parameters_map["power"]["points"]["data_sheet_units"],
                ).to("kW")
            else:
                power_shaft_guarantee = Q_(
                    float(st.session_state["power_shaft_point_guarantee"]),
                    parameters_map["power_shaft"]["points"]["data_sheet_units"],
                ).to("kW")
            kwargs_guarantee["power_losses"] = power_shaft_guarantee - power_guarantee
        else:
            kwargs_guarantee["power_losses"] = Q_(0, "W")

        time.sleep(0.1)
        progress_value += 5
        progress_bar.progress(progress_value, text="Calculating guarantee point...")
        guarantee_point = ccp.Point(
            **kwargs_guarantee,
        )

        for i in range(1, number_of_test_points + 1):
            kwargs = {}
            # check if at least flow, suction pressure and suction temperature are filled
            if (
                st.session_state[f"flow_point_{i}"] == ""
                or st.session_state[f"suction_pressure_point_{i}"] == ""
                or st.session_state[f"suction_temperature_point_{i}"] == ""
            ):
                if i < 6:
                    st.warning(f"Please fill the data for point {i}")
                continue
            else:
                calculate_button = False
                if (
                    Q_(0, parameters_map["flow"]["points"]["test_units"]).dimensionality
                    == "[mass] / [time]"
                ):
                    if st.session_state[f"flow_point_{i}"]:
                        kwargs["flow_m"] = Q_(
                            float(st.session_state[f"flow_point_{i}"]),
                            parameters_map["flow"]["points"]["test_units"],
                        )
                    else:
                        kwargs["flow_m"] = None
                else:
                    if st.session_state[f"flow_point_{i}"]:
                        kwargs["flow_v"] = Q_(
                            float(st.session_state[f"flow_point_{i}"]),
                            parameters_map["flow"]["points"]["test_units"],
                        )
                    else:
                        kwargs["flow_v"] = None
                kwargs["suc"] = ccp.State(
                    p=Q_(
                        float(st.session_state[f"suction_pressure_point_{i}"]),
                        parameters_map["suction_pressure"]["points"]["test_units"],
                    ),
                    T=Q_(
                        float(st.session_state[f"suction_temperature_point_{i}"]),
                        parameters_map["suction_temperature"]["points"]["test_units"],
                    ),
                    fluid=get_gas_composition(
                        st.session_state[f"gas_point_{i}"],
                        gas_compositions_table,
                        default_components,
                    ),
                )
                kwargs["disch"] = ccp.State(
                    p=Q_(
                        float(st.session_state[f"discharge_pressure_point_{i}"]),
                        parameters_map["discharge_pressure"]["points"]["test_units"],
                    ),
                    T=Q_(
                        float(st.session_state[f"discharge_temperature_point_{i}"]),
                        parameters_map["discharge_temperature"]["points"]["test_units"],
                    ),
                    fluid=get_gas_composition(
                        st.session_state[f"gas_point_{i}"],
                        gas_compositions_table,
                        default_components,
                    ),
                )
                if st.session_state[f"casing_delta_T_point_{i}"] and casing_heat_loss:
                    kwargs["casing_temperature"] = Q_(
                        float(st.session_state[f"casing_delta_T_point_{i}"]),
                        parameters_map["casing_delta_T"]["points"]["test_units"],
                    )
                    kwargs["ambient_temperature"] = 0
                else:
                    kwargs["casing_temperature"] = 0
                    kwargs["ambient_temperature"] = 0

                if calculate_leakages:
                    if st.session_state[f"balance_line_flow_m_point_{i}"] != "":
                        kwargs["balance_line_flow_m"] = Q_(
                            float(st.session_state[f"balance_line_flow_m_point_{i}"]),
                            parameters_map["balance_line_flow_m"]["points"][
                                "test_units"
                            ],
                        )
                    else:
                        kwargs["balance_line_flow_m"] = None
                    if (
                        seal_gas_flow
                        and st.session_state[f"seal_gas_flow_m_point_{i}"] != ""
                    ):
                        kwargs["seal_gas_flow_m"] = Q_(
                            float(st.session_state[f"seal_gas_flow_m_point_{i}"]),
                            parameters_map["seal_gas_flow_m"]["points"]["test_units"],
                        )
                    else:
                        kwargs["seal_gas_flow_m"] = Q_(
                            0, parameters_map["seal_gas_flow_m"]["points"]["test_units"]
                        )
                    if (
                        seal_gas_flow
                        and st.session_state[f"seal_gas_temperature_point_{i}"] != ""
                    ):
                        kwargs["seal_gas_temperature"] = Q_(
                            float(st.session_state[f"seal_gas_temperature_point_{i}"]),
                            parameters_map["seal_gas_temperature"]["points"][
                                "test_units"
                            ],
                        )
                    else:
                        kwargs["seal_gas_temperature"] = Q_(
                            0,
                            parameters_map["seal_gas_temperature"]["points"][
                                "test_units"
                            ],
                        )
                else:
                    pass
                    # TODO implement calculation without leakages

                kwargs["bearing_mechanical_losses"] = bearing_mechanical_losses
                if bearing_mechanical_losses:
                    if st.session_state[f"oil_flow_journal_bearing_de_point_{i}"] != "":
                        kwargs["oil_flow_journal_bearing_de"] = Q_(
                            float(
                                st.session_state[
                                    f"oil_flow_journal_bearing_de_point_{i}"
                                ]
                            ),
                            parameters_map["oil_flow_journal_bearing_de"]["points"][
                                "test_units"
                            ],
                        )
                    else:
                        kwargs["oil_flow_journal_bearing_nde"] = None
                    if (
                        st.session_state[f"oil_flow_journal_bearing_nde_point_{i}"]
                        != ""
                    ):
                        kwargs["oil_flow_journal_bearing_nde"] = Q_(
                            float(
                                st.session_state[
                                    f"oil_flow_journal_bearing_nde_point_{i}"
                                ]
                            ),
                            parameters_map["oil_flow_journal_bearing_nde"]["points"][
                                "test_units"
                            ],
                        )
                    else:
                        kwargs["oil_flow_thrust_bearing_nde"] = None
                    if st.session_state[f"oil_flow_thrust_bearing_nde_point_{i}"] != "":
                        kwargs["oil_flow_thrust_bearing_nde"] = Q_(
                            float(
                                st.session_state[
                                    f"oil_flow_thrust_bearing_nde_point_{i}"
                                ]
                            ),
                            parameters_map["oil_flow_thrust_bearing_nde"]["points"][
                                "test_units"
                            ],
                        )
                    else:
                        kwargs["oil_inlet_temperature"] = None
                    if st.session_state[f"oil_inlet_temperature_point_{i}"] != "":
                        kwargs["oil_inlet_temperature"] = Q_(
                            float(st.session_state[f"oil_inlet_temperature_point_{i}"]),
                            parameters_map["oil_inlet_temperature"]["points"][
                                "test_units"
                            ],
                        )
                    else:
                        kwargs["oil_outlet_temperature_de"] = None
                    if st.session_state[f"oil_outlet_temperature_de_point_{i}"] != "":
                        kwargs["oil_outlet_temperature_de"] = Q_(
                            float(
                                st.session_state[f"oil_outlet_temperature_de_point_{i}"]
                            ),
                            parameters_map["oil_outlet_temperature_de"]["points"][
                                "test_units"
                            ],
                        )
                    else:
                        kwargs["oil_outlet_temperature_nde"] = None
                    if st.session_state[f"oil_outlet_temperature_nde_point_{i}"] != "":
                        kwargs["oil_outlet_temperature_nde"] = Q_(
                            float(
                                st.session_state[
                                    f"oil_outlet_temperature_nde_point_{i}"
                                ]
                            ),
                            parameters_map["oil_outlet_temperature_nde"]["points"][
                                "test_units"
                            ],
                        )
                    else:
                        kwargs["oil_outlet_temperature_nde"] = None
                if oil_iso and bearing_mechanical_losses:
                    oil_specific_heat_de = specific_heat_calculate(
                        kwargs["oil_inlet_temperature"],
                        kwargs["oil_outlet_temperature_de"],
                        oil_iso_classification,
                    )
                    oil_specific_heat_nde = specific_heat_calculate(
                        kwargs["oil_inlet_temperature"],
                        kwargs["oil_outlet_temperature_nde"],
                        oil_iso_classification,
                    )

                    oil_density_de = density_calculate(
                        kwargs["oil_inlet_temperature"],
                        kwargs["oil_outlet_temperature_de"],
                        oil_iso_classification,
                    )
                    oil_density_nde = density_calculate(
                        kwargs["oil_inlet_temperature"],
                        kwargs["oil_outlet_temperature_nde"],
                        oil_iso_classification,
                    )

                    kwargs["oil_specific_heat_de"] = oil_specific_heat_de
                    kwargs["oil_specific_heat_nde"] = oil_specific_heat_nde

                    kwargs["oil_density_de"] = oil_density_de
                    kwargs["oil_density_nde"] = oil_density_nde

                elif oil_specific_heat and bearing_mechanical_losses:
                    kwargs["oil_specific_heat_de"] = oil_specific_heat_value
                    kwargs["oil_specific_heat_nde"] = oil_specific_heat_value
                    kwargs["oil_density_de"] = oil_density_value
                    kwargs["oil_density_nde"] = oil_density_value

                kwargs["b"] = Q_(
                    float(st.session_state["b_point_guarantee"]),
                    parameters_map["b"]["points"]["data_sheet_units"],
                )
                kwargs["D"] = Q_(
                    float(st.session_state["D_point_guarantee"]),
                    parameters_map["D"]["points"]["data_sheet_units"],
                )
                kwargs["casing_area"] = Q_(
                    float(st.session_state["casing_area_point_guarantee"]),
                    parameters_map["casing_area"]["points"]["data_sheet_units"],
                )
                kwargs["surface_roughness"] = Q_(
                    float(st.session_state["surface_roughness_point_guarantee"]),
                    parameters_map["surface_roughness"]["points"]["data_sheet_units"],
                )
                time.sleep(0.1)
                progress_value += 2
                progress_bar.progress(progress_value, text="Calculating test points...")

                test_points.append(
                    Point1Sec(
                        speed=Q_(
                            float(st.session_state[f"speed_point_{i}"]),
                            parameters_map["speed"]["points"]["test_units"],
                        ),
                        **kwargs,
                    )
                )

        time.sleep(0.1)
        progress_value += 10
        progress_bar.progress(progress_value, text="Converting points...")
        straight_through = StraightThrough(
            guarantee_point=guarantee_point,
            test_points=test_points,
            reynolds_correction=reynolds_correction,
            bearing_mechanical_losses=bearing_mechanical_losses,
        )

        if calculate_speed_button:
            time.sleep(0.1)
            progress_value += 10
            progress_bar.progress(progress_value, text="Finding speed...")
            straight_through = (
                straight_through.calculate_speed_to_match_discharge_pressure()
            )

        time.sleep(0.1)
        progress_bar.progress(100, text="Done!")

        # add straight_through object to session state
        st.session_state["straight_through"] = straight_through

    # if straight_through is not defined, pickle the saved file
    if (
        st.session_state["straight_through"] is not None
        and st.session_state["straight_through"] != ""
    ):
        straight_through = st.session_state["straight_through"]

    if (
        st.session_state["straight_through"] is not None
        and st.session_state["straight_through"] != ""
    ):
        with st.expander("Results"):
            st.write(
                f"Final speed(s) used in calculation: {straight_through.speed_operational.to('rpm').m:.2f} RPM"
            )

            _t = "\u209c"
            _sp = "\u209b\u209a"
            conv = "\u1d9c" + "\u1d52" + "\u207f" + "\u1d5b"

            results = {}

            if straight_through:
                # create interpolated point with point method
                point_interpolated = getattr(straight_through, "point")(
                    flow_v=getattr(straight_through, "guarantee_point").flow_v,
                    speed=straight_through.speed_operational,
                )

                results[f"φ{_t}"] = [
                    round(p.phi.m, 5) for p in getattr(straight_through, "test_points")
                ]
                results[f"φ{_t} / φ{_sp}"] = [
                    round(
                        p.phi.m / getattr(straight_through, "guarantee_point").phi.m,
                        5,
                    )
                    for p in getattr(straight_through, "test_points")
                ]
                results["vi / vd"] = [
                    round(p.volume_ratio.m, 5)
                    for p in getattr(straight_through, "test_points")
                ]
                results[f"(vi/vd){_t}/(vi/vd){_sp}"] = [
                    round(
                        p.volume_ratio.m
                        / getattr(straight_through, "guarantee_point").volume_ratio.m,
                        5,
                    )
                    for p in getattr(straight_through, "test_points")
                ]
                results[f"Mach{_t}"] = [
                    round(p.mach.m, 5) for p in getattr(straight_through, "test_points")
                ]
                results[f"Mach{_t} - Mach{_sp}"] = [
                    round(
                        p.mach.m - getattr(straight_through, "guarantee_point").mach.m,
                        5,
                    )
                    for p in getattr(straight_through, "test_points")
                ]
                results[f"Re{_t}"] = [
                    round(p.reynolds.m, 5)
                    for p in getattr(straight_through, "test_points")
                ]
                results[f"Re{_t} / Re{_sp}"] = [
                    round(
                        p.reynolds.m
                        / getattr(straight_through, "guarantee_point").reynolds.m,
                        5,
                    )
                    for p in getattr(straight_through, "test_points")
                ]
                results[f"pd{conv} (bar)"] = [
                    round(p.disch.p("bar").m, 5)
                    for p in getattr(straight_through, "points_flange_sp")
                ]
                results[f"pd{conv}/pd{_sp}"] = [
                    round(
                        p.disch.p("bar").m
                        / getattr(straight_through, "guarantee_point").disch.p("bar").m,
                        5,
                    )
                    for p in getattr(straight_through, "points_flange_sp")
                ]
                results[f"Head{_t} (kJ/kg)"] = [
                    round(p.head.to("kJ/kg").m, 5)
                    for p in getattr(straight_through, "points_flange_sp")
                ]
                results[f"Head{_t}/Head{_sp}"] = [
                    round(
                        p.head.to("kJ/kg").m
                        / getattr(straight_through, "guarantee_point")
                        .head.to("kJ/kg")
                        .m,
                        5,
                    )
                    for p in getattr(straight_through, "test_points")
                ]
                results[f"Head{conv} (kJ/kg)"] = [
                    round(p.head.to("kJ/kg").m, 5)
                    for p in getattr(straight_through, "points_flange_sp")
                ]
                results[f"Head{conv}/Head{_sp}"] = [
                    round(
                        p.head.to("kJ/kg").m
                        / getattr(straight_through, "guarantee_point")
                        .head.to("kJ/kg")
                        .m,
                        5,
                    )
                    for p in getattr(straight_through, "points_flange_sp")
                ]
                results[f"Q{conv} (m3/h)"] = [
                    round(p.flow_v.to("m³/h").m, 5)
                    for p in getattr(straight_through, "points_flange_sp")
                ]
                results[f"Q{conv}/Q{_sp}"] = [
                    round(
                        p.flow_v.to("m³/h").m
                        / getattr(straight_through, "guarantee_point")
                        .flow_v.to("m³/h")
                        .m,
                        5,
                    )
                    for p in getattr(straight_through, "points_flange_sp")
                ]
                results[f"W{_t} (kW)"] = [
                    round(p.power_shaft.to("kW").m, 5)
                    for p in getattr(straight_through, "points_rotor_t")
                ]
                if bearing_mechanical_losses:
                    results[f"W{_t}/W{_sp}"] = [
                        round(
                            p.power_shaft.to("kW").m
                            / Q_(
                                float(st.session_state["power_shaft_point_guarantee"]),
                                parameters_map["power_shaft"]["points"][
                                    "data_sheet_units"
                                ],
                            )
                            .to("kW")
                            .m,
                            5,
                        )
                        for p in getattr(straight_through, "points_rotor_t")
                    ]
                else:
                    results[f"W{_t}/W{_sp}"] = [
                        round(
                            p.power_shaft.to("kW").m
                            / Q_(
                                float(st.session_state["power_point_guarantee"]),
                                parameters_map["power"]["points"]["data_sheet_units"],
                            )
                            .to("kW")
                            .m,
                            5,
                        )
                        for p in getattr(straight_through, "points_rotor_t")
                    ]
                results[f"W{conv} (kW)"] = [
                    round(p.power_shaft.to("kW").m, 5)
                    for p in getattr(straight_through, "points_rotor_sp")
                ]
                if bearing_mechanical_losses:
                    results[f"W{conv}/W{_sp}"] = [
                        round(
                            p.power_shaft.to("kW").m
                            / Q_(
                                float(st.session_state["power_shaft_point_guarantee"]),
                                parameters_map["power_shaft"]["points"][
                                    "data_sheet_units"
                                ],
                            )
                            .to("kW")
                            .m,
                            5,
                        )
                        for p in getattr(straight_through, "points_rotor_sp")
                    ]
                else:
                    results[f"W{conv}/W{_sp}"] = [
                        round(
                            p.power_shaft.to("kW").m
                            / Q_(
                                float(st.session_state["power_point_guarantee"]),
                                parameters_map["power"]["points"]["data_sheet_units"],
                            )
                            .to("kW")
                            .m,
                            5,
                        )
                        for p in getattr(straight_through, "points_rotor_sp")
                    ]
                results[f"Eff{_t}"] = [
                    round(p.eff.m, 5)
                    for p in getattr(straight_through, "points_flange_t")
                ]
                results[f"Eff{conv}"] = [
                    round(p.eff.m, 5)
                    for p in getattr(straight_through, "points_flange_sp")
                ]

                results[f"φ{_t}"].append(round(point_interpolated.phi.m, 5))
                results[f"φ{_t} / φ{_sp}"].append(
                    round(
                        point_interpolated.phi.m
                        / getattr(straight_through, "guarantee_point").phi.m,
                        5,
                    )
                )
                results["vi / vd"].append(round(point_interpolated.volume_ratio.m, 5))
                results[f"(vi/vd){_t}/(vi/vd){_sp}"].append(
                    round(
                        point_interpolated.volume_ratio.m
                        / getattr(straight_through, "guarantee_point").volume_ratio.m,
                        5,
                    )
                )
                results[f"Mach{_t}"].append(round(point_interpolated.mach.m, 5))
                results[f"Mach{_t} - Mach{_sp}"].append(
                    round(
                        point_interpolated.mach.m
                        - getattr(straight_through, "guarantee_point").mach.m,
                        5,
                    )
                )
                results[f"Re{_t}"].append(round(point_interpolated.reynolds.m, 5))
                results[f"Re{_t} / Re{_sp}"].append(
                    round(
                        point_interpolated.reynolds.m
                        / getattr(straight_through, "guarantee_point").reynolds.m,
                        5,
                    )
                )
                results[f"pd{conv} (bar)"].append(
                    round(point_interpolated.disch.p("bar").m, 5)
                )
                results[f"pd{conv}/pd{_sp}"].append(
                    round(
                        point_interpolated.disch.p("bar").m
                        / getattr(straight_through, "guarantee_point").disch.p("bar").m,
                        5,
                    )
                )
                results[f"Head{_t} (kJ/kg)"].append(
                    round(point_interpolated.head.to("kJ/kg").m, 5)
                )
                results[f"Head{_t}/Head{_sp}"].append(
                    round(
                        point_interpolated.head.to("kJ/kg").m
                        / getattr(straight_through, "guarantee_point")
                        .head.to("kJ/kg")
                        .m,
                        5,
                    )
                )
                results[f"Head{conv} (kJ/kg)"].append(
                    round(point_interpolated.head.to("kJ/kg").m, 5)
                )
                results[f"Head{conv}/Head{_sp}"].append(
                    round(
                        point_interpolated.head.to("kJ/kg").m
                        / getattr(straight_through, "guarantee_point")
                        .head.to("kJ/kg")
                        .m,
                        5,
                    )
                )
                results[f"Q{conv} (m3/h)"].append(
                    round(point_interpolated.flow_v.to("m³/h").m, 5)
                )
                results[f"Q{conv}/Q{_sp}"].append(
                    round(
                        point_interpolated.flow_v.to("m³/h").m
                        / getattr(straight_through, "guarantee_point")
                        .flow_v.to("m³/h")
                        .m,
                        5,
                    )
                )
                results[f"W{_t} (kW)"].append(None)
                results[f"W{_t}/W{_sp}"].append(None)
                results[f"W{conv} (kW)"].append(
                    round(point_interpolated.power_shaft.to("kW").m, 5)
                )
                if bearing_mechanical_losses:
                    results[f"W{conv}/W{_sp}"].append(
                        round(
                            point_interpolated.power_shaft.to("kW").m
                            / Q_(
                                float(st.session_state["power_shaft_point_guarantee"]),
                                parameters_map["power_shaft"]["points"][
                                    "data_sheet_units"
                                ],
                            )
                            .to("kW")
                            .m,
                            5,
                        )
                    )
                else:
                    results[f"W{conv}/W{_sp}"].append(
                        round(
                            point_interpolated.power_shaft.to("kW").m
                            / Q_(
                                float(st.session_state["power_point_guarantee"]),
                                parameters_map["power"]["points"]["data_sheet_units"],
                            )
                            .to("kW")
                            .m,
                            5,
                        )
                    )
                results[f"Eff{_t}"].append(round(point_interpolated.eff.m, 5))
                results[f"Eff{conv}"].append(round(point_interpolated.eff.m, 5))

                df_results = pd.DataFrame(results)
                rename_index = {
                    i: f"Point {i+1}"
                    for i in range(len(getattr(straight_through, "points_flange_t")))
                }
                rename_index[len(getattr(straight_through, "points_flange_t"))] = (
                    "Guarantee Point"
                )

                df_results.rename(
                    index=rename_index,
                    inplace=True,
                )

                def highlight_cell(
                    styled_df, df, row_index, col_index, lower_limit, higher_limit
                ):
                    """Applies conditional formatting to a specific cell in a pandas DataFrame.

                    Args:
                        df (pandas.DataFrame): The DataFrame to apply conditional formatting to.
                        row_index (int or str): The row index of the cell to highlight.
                        col_index (int or str): The column index of the cell to highlight.
                        lower_limit (float): The lower limit of the value range to highlight in green.
                        higher_limit (float): The higher limit of the value range to highlight in green.

                    Returns:
                        pandas.io.formats.style.Styler: A styled DataFrame with the specified cell highlighted.
                    """
                    # create a copy of the DataFrame with styling

                    # apply conditional formatting to the specific cell
                    cell_value = df.loc[row_index, col_index]
                    if cell_value >= lower_limit and cell_value <= higher_limit:
                        styled_df = styled_df.map(
                            lambda x: (
                                "background-color: #C8E6C9" if x == cell_value else ""
                            )
                        ).map(
                            lambda x: "font-color: #33691E" if x == cell_value else ""
                        )

                    else:
                        styled_df = styled_df.map(
                            lambda x: (
                                "background-color: #FFCDD2" if x == cell_value else ""
                            )
                        ).map(
                            lambda x: "font-color: #FFCDD2" if x == cell_value else ""
                        )

                    return styled_df

                styled_df_results = df_results.style

                mach_limits = point_interpolated.mach_limits()
                styled_df_results = highlight_cell(
                    styled_df_results,
                    df_results,
                    "Guarantee Point",
                    f"Mach{_t} - Mach{_sp}",
                    mach_limits["lower"],
                    mach_limits["upper"],
                )
                reynolds_limits = point_interpolated.reynolds_limits()
                styled_df_results = highlight_cell(
                    styled_df_results,
                    df_results,
                    "Guarantee Point",
                    f"Re{_t} / Re{_sp}",
                    reynolds_limits["lower"],
                    reynolds_limits["upper"],
                )
                styled_df_results = highlight_cell(
                    styled_df_results,
                    df_results,
                    "Guarantee Point",
                    f"(vi/vd){_t}/(vi/vd){_sp}",
                    0.95,
                    1.05,
                )

                if variable_speed:
                    power_limit = 1.04
                    pressure_limit = 1e15
                else:
                    power_limit = 1.07
                    pressure_limit = 1.05

                for _point in ["Guarantee Point"]:
                    # highlight pdconv/pdsp
                    styled_df_results = highlight_cell(
                        styled_df_results,
                        df_results,
                        _point,
                        f"pd{conv}/pd{_sp}",
                        1.0,
                        pressure_limit,
                    )
                    # highlight Wconv/Wsp
                    styled_df_results = highlight_cell(
                        styled_df_results,
                        df_results,
                        _point,
                        f"W{conv}/W{_sp}",
                        0.0,
                        power_limit,
                    )

                st.dataframe(
                    styled_df_results.format(
                        "{:.2%}",
                        subset=[
                            f"φ{_t} / φ{_sp}",
                            f"pd{conv}/pd{_sp}",
                            f"(vi/vd){_t}/(vi/vd){_sp}",
                            f"Head{_t}/Head{_sp}",
                            f"Head{conv}/Head{_sp}",
                            f"Q{conv}/Q{_sp}",
                            f"W{_t}/W{_sp}",
                            f"W{conv}/W{_sp}",
                            f"Eff{_t}",
                            f"Eff{conv}",
                        ],
                    ).format(
                        "{:.4e}",
                        subset=[
                            f"Re{_t}",
                        ],
                    )
                )

                table_excel = to_excel(styled_df_results)
                st.download_button(
                    "Download Results Table",
                    data=table_excel,
                    file_name="results.xlsx",
                    mime="application/vnd.ms-excel",
                    use_container_width=True,
                )

                with st.container():
                    mach_col, reynolds_col = st.columns(2)
                    mach_col.plotly_chart(
                        point_interpolated.plot_mach(), use_container_width=True
                    )
                    reynolds_col.plotly_chart(
                        point_interpolated.plot_reynolds(), use_container_width=True
                    )

                def add_background_image(curve_name=None, fig=None, image=None):
                    """Add png file to plot background

                    Parameters
                    ----------
                    curve_name : str
                        The name of the curve to add the background image to.
                    fig : plotly.graph_objects.Figure
                        The figure to add the background image to.
                    image : io.BytesIO
                        The image to add to the background.
                    """
                    encoded_string = base64.b64encode(image).decode()
                    encoded_image = "data:image/png;base64," + encoded_string
                    fig.add_layout_image(
                        dict(
                            source=encoded_image,
                            xref="x",
                            yref="y",
                            x=plot_limits[curve_name]["x"]["lower_limit"],
                            y=plot_limits[curve_name]["y"]["upper_limit"],
                            sizex=float(plot_limits[curve_name]["x"]["upper_limit"])
                            - float(plot_limits[curve_name]["x"]["lower_limit"]),
                            sizey=float(plot_limits[curve_name]["y"]["upper_limit"])
                            - float(plot_limits[curve_name]["y"]["lower_limit"]),
                            sizing="stretch",
                            opacity=0.5,
                            layer="below",
                        )
                    )
                    return fig

                plots_dict = {}
                for curve in ["head", "eff", "discharge_pressure", "power"]:
                    flow_v_units = plot_limits.get(curve, {}).get("x", {}).get("units")
                    curve_units = plot_limits.get(curve, {}).get("y", {}).get("units")

                    kwargs = {}
                    if flow_v_units is not None and flow_v_units != "":
                        kwargs["flow_v_units"] = flow_v_units
                    if curve_units is not None and curve_units != "":
                        if curve == "discharge_pressure":
                            kwargs["p_units"] = curve_units
                        else:
                            kwargs[f"{curve}_units"] = curve_units

                    if curve == "discharge_pressure":
                        curve_plot_method = "disch.p"
                    else:
                        curve_plot_method = curve

                    plots_dict[curve] = r_getattr(
                        straight_through,
                        # getattr(straight_through, f"imp_flange_sp"),
                        f"{curve_plot_method}_plot",
                    )(
                        show_points=show_points,
                        **kwargs,
                    )
                    plots_dict[curve] = r_getattr(
                        point_interpolated, f"{curve_plot_method}_plot"
                    )(
                        fig=plots_dict[curve],
                        show_points=show_points,
                        **kwargs,
                    )

                    plots_dict[curve].data[0].update(
                        name=f"Converted Curve {straight_through.speed_operational.to('rpm').m:.0f} RPM",
                    )
                    if curve == "discharge_pressure":
                        plots_dict[curve].data[1].update(
                            name=f"Flow: {point_interpolated.flow_v.to(flow_v_units):.~2f}, {curve.capitalize()}: {r_getattr(point_interpolated, curve_plot_method)(curve_units):.~2f}".replace(
                                "m ** 3 / h", "m³/h"
                            ).replace(
                                "Discharge_pressure", "Disch. p"
                            )
                        )
                    else:
                        plots_dict[curve].data[1].update(
                            name=f"Flow: {point_interpolated.flow_v.to(flow_v_units):.~2f}, {curve.capitalize()}: {r_getattr(point_interpolated, curve_plot_method).to(curve_units):.~2f}".replace(
                                "m ** 3 / h", "m³/h"
                            ).replace(
                                "Discharge_pressure", "Disch. p"
                            )
                        )

                    plots_dict[curve].update_layout(
                        showlegend=True,
                        # position legend at the bottom left
                        legend=dict(
                            yanchor="bottom",
                            y=0.01,
                            xanchor="left",
                            x=0.01,
                        ),
                    )

                    x_lower = plot_limits.get(curve, {}).get("x", {}).get("lower_limit")
                    x_upper = plot_limits.get(curve, {}).get("x", {}).get("upper_limit")
                    y_lower = plot_limits.get(curve, {}).get("y", {}).get("lower_limit")
                    y_upper = plot_limits.get(curve, {}).get("y", {}).get("upper_limit")

                    if (
                        x_lower is not None
                        and x_lower != ""
                        and x_upper is not None
                        and x_upper != ""
                        and y_lower is not None
                        and y_lower != ""
                        and y_upper is not None
                        and y_upper != ""
                    ):
                        plots_dict[curve].update_layout(
                            xaxis_range=(
                                plot_limits[curve]["x"]["lower_limit"],
                                plot_limits[curve]["x"]["upper_limit"],
                            ),
                            yaxis_range=(
                                plot_limits[curve]["y"]["lower_limit"],
                                plot_limits[curve]["y"]["upper_limit"],
                            ),
                        )

                    if (
                        st.session_state.get(f"fig_{curve}") is not None
                        and st.session_state.get(f"fig_{curve}") != ""
                    ):
                        plots_dict[curve] = add_background_image(
                            curve_name=curve,
                            fig=plots_dict[curve],
                            image=st.session_state[f"fig_{curve}"],
                        )

                with st.container():
                    head_col, eff_col = st.columns(2)
                    head_col.plotly_chart(plots_dict["head"], use_container_width=True)
                    eff_col.plotly_chart(plots_dict["eff"], use_container_width=True)

                with st.container():
                    disch_p_col, power_col = st.columns(2)
                    disch_p_col.plotly_chart(
                        plots_dict["discharge_pressure"], use_container_width=True
                    )
                    power_col.plotly_chart(
                        plots_dict["power"], use_container_width=True
                    )


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        logging.info("app: straigh_through")
        logging.info(f"session state: {st.session_state}")
        logging.error(e)
        raise e



"""IO module.

yaml will be the base format to IO. Other file types will be converted to yaml
before beeing loaded.

"""

from .processing import fluctuation, fluctuation_data, mean_data, filter_data



"""Module to define compressors with 1 or 2 sections."""

from copy import copy

import ccp
import toml
from ccp.impeller import Impeller
from ccp.point import Point, flow_from_phi
from ccp.state import State
from ccp.config.units import check_units
from ccp import Q_
import numpy as np
from scipy.optimize import newton


class Point1Sec(Point):
    """Point class for a compressor with 1 section.

    Parameters
    ----------
    balance_line_flow_m : float, pint.Quantity
        Balance line mass flow (kg/s).
    seal_gas_flow_m : float, pint.Quantity
        Seal gas mass flow (kg/s).
    seal_gas_temperature : float, pint.Quantity
        Seal gas injection temperature (degK).
    oil_flow_journal_bearing_de : float, pint.Quantity
        Oil flow journal bearing drive end side (m³/s).
    oil_flow_journal_bearing_nde : float, pint.Quantity
        Oil flow journal bearing non-drive end side (m³/s).
    oil_flow_thrust_bearing_nde : float, pint.Quantity
         Oil flow journal bearing thrust bearing (m³/s).
    oil_inlet_temperature : float, pint.Quantity
        Oil inlet temperature (degK).
    oil_outlet_temperature_de : float, pint.Quantity
        Oil outlet temperature journal bearing drive end side (degK).
    oil_outlet_temperature_nde : float, pint.Quantity
        Oil outlet temperature bearing non-drive end side (degK).
    oil_specific_heat_de : float, pint.Quantity
        Oil specific heat bearing drive end side (J/kg/degK).
    oil_specific_heat_nde : float, pint.Quantity
        Oil specific heat bearing non-drive end side (J/kg/degK).
    oil_density_de : float, pint.Quantity
        Oil Density bearing drive end side (kg/m³).
    oil_density_nde : float, pint.Quantity
        Oil density bearing non-drive end side (kg/m³).

    Returns
    -------
    point1sec : ccp.Point1Sec
        A point for a straight through compressor (inherited from ccp.Point).

    Examples
    --------
    >>> import ccp
    >>> p = Point1Sec(
    ...     flow_m=Q_(7.737, "kg/s"),
    ...     speed=Q_(7894, "RPM"),
    ...     b=Q_(28.5, "mm"),
    ...     D=Q_(365, "mm"),
    ...     suc=State(
    ...         p=Q_(1.826, "bar"),
    ...         T=Q_(296.7, "degK"),
    ...         fluid={
    ...             "carbon dioxide": 0.80218,
    ...             "R134a": 0.18842,
    ...             "nitrogen": 0.0091,
    ...             "oxygen": 0.0003,
    ...         },
    ...     ),
    ...     disch=State(
    ...         p=Q_(6.142, "bar"),
    ...         T=Q_(392.1, "degK"),
    ...         fluid={
    ...             "carbon dioxide": 0.80218,
    ...             "R134a": 0.18842,
    ...             "nitrogen": 0.0091,
    ...             "oxygen": 0.0003,
    ...         },
    ...     ),
    ...     balance_line_flow_m=Q_(0.1076, "kg/s"),
    ...     seal_gas_flow_m=Q_(0.04982, "kg/s"),
    ...     seal_gas_temperature=Q_(297.7, "degK"),
    ...     oil_flow_journal_bearing_de=Q_(27.084, "l/min"),
    ...     oil_flow_journal_bearing_nde=Q_(47.984, "l/min"),
    ...     oil_flow_thrust_bearing_nde=Q_(33.52, "l/min"),
    ...     oil_inlet_temperature=Q_(42.184, "degC"),
    ...     oil_outlet_temperature_de=Q_(48.111, "degC"),
    ...     oil_outlet_temperature_nde=Q_(46.879, "degC"),
    ...     oil_specific_heat_de=Q_(2.02, "kJ/kg/degK"),
    ...     oil_specific_heat_nde=Q_(2.02, "kJ/kg/degK"),
    ...     oil_density_de=Q_(846.9, "kg/m³"),
    ...     oil_density_nde=Q_(846.9, "kg/m³"),
    ...     casing_area=7.5,
    ...     casing_temperature=Q_(31.309, "degC"),
    ...     ambient_temperature=Q_(0, "degC"),
    ...     )

    """

    @check_units
    def __init__(
        self,
        *args,
        balance_line_flow_m=None,
        seal_gas_flow_m=None,
        seal_gas_temperature=None,
        oil_flow_journal_bearing_de=None,
        oil_flow_journal_bearing_nde=None,
        oil_flow_thrust_bearing_nde=None,
        oil_inlet_temperature=None,
        oil_outlet_temperature_de=None,
        oil_outlet_temperature_nde=None,
        oil_specific_heat_de=None,
        oil_specific_heat_nde=None,
        oil_density_de=None,
        oil_density_nde=None,
        leakages=True,
        bearing_mechanical_losses=True,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )
        self.balance_line_flow_m = balance_line_flow_m
        self.seal_gas_flow_m = seal_gas_flow_m
        self.seal_gas_temperature = seal_gas_temperature
        self.oil_flow_journal_bearing_de = oil_flow_journal_bearing_de
        self.oil_flow_journal_bearing_nde = oil_flow_journal_bearing_nde
        self.oil_flow_thrust_bearing_nde = oil_flow_thrust_bearing_nde
        self.oil_inlet_temperature = oil_inlet_temperature
        self.oil_outlet_temperature_de = oil_outlet_temperature_de
        self.oil_outlet_temperature_nde = oil_outlet_temperature_nde
        self.oil_specific_heat_de = oil_specific_heat_de
        self.oil_specific_heat_nde = oil_specific_heat_nde
        self.oil_density_de = oil_density_de
        self.oil_density_nde = oil_density_nde
        self.leakages = leakages
        self.bearing_mechanical_losses = bearing_mechanical_losses

        if not self.leakages:
            self.balance_line_flow_m = Q_(0, "kg/s")
            self.seal_gas_flow_m = Q_(0, "kg/s")
            self.seal_gas_temperature = self.suc.T()

        if not self.bearing_mechanical_losses:
            self.oil_flow_journal_bearing_de = Q_(0, "m³/s")
            self.oil_flow_journal_bearing_nde = Q_(0, "m³/s")
            self.oil_flow_thrust_bearing_nde = Q_(0, "m³/s")
            self.oil_inlet_temperature = self.suc.T()
            self.oil_outlet_temperature_de = self.suc.T()
            self.oil_outlet_temperature_nde = self.suc.T()
            oil_specific_heat_de = (Q_(2.02, "kJ/kg/degK"),)
            oil_specific_heat_nde = (Q_(2.02, "kJ/kg/degK"),)
            oil_density_de = (Q_(846.9, "kg/m³"),)
            oil_density_nde = (Q_(846.9, "kg/m³"),)

    def _dict_to_save(self):
        """Returns a dict that will be saved to a toml file."""
        dict_to_save = super()._dict_to_save()
        for param in [
            "balance_line_flow_m",
            "seal_gas_flow_m",
            "seal_gas_temperature",
            "oil_flow_journal_bearing_de",
            "oil_flow_journal_bearing_nde",
            "oil_flow_thrust_bearing_nde",
            "oil_inlet_temperature",
            "oil_outlet_temperature_de",
            "oil_outlet_temperature_nde",
            "oil_specific_heat_de",
            "oil_specific_heat_nde",
            "oil_density_de",
            "oil_density_nde",
        ]:
            if getattr(self, param):
                dict_to_save[param] = str(getattr(self, param))

        return dict_to_save


class StraightThrough(Impeller):
    """Straight Through compressor"""

    @check_units
    def __init__(
        self,
        guarantee_point,
        test_points,
        speed_operational=None,
        reynolds_correction=False,
        bearing_mechanical_losses=False,
    ):
        self.guarantee_point = guarantee_point
        self.test_points = test_points
        if speed_operational is None:
            speed_operational = guarantee_point.speed
        self.speed_operational = speed_operational
        self.reynolds_correction = reynolds_correction
        self.bearing_mechanical_losses = bearing_mechanical_losses

        # points for test flange conditions
        self.points_flange_t = test_points

        # calculate rotor condition
        test_points_rotor = []
        self.k_end_seal = []  # list with seal constants
        for point in test_points:
            ms1f = point.flow_m
            mbal = point.balance_line_flow_m
            mseal = point.seal_gas_flow_m
            if mbal == None:
                mbal = Q_(0, "kg/s")
            if mseal == None:
                mseal = Q_(0, "kg/s")

            mend = mbal - (0.95 * mseal) / 2
            ms1r = ms1f + mend

            Ts1f = point.suc.T()
            # dummy state to calculate Tend
            dummy_state = copy(point.disch)
            dummy_state.update(p=point.suc.p(), h=dummy_state.h())
            Tend = dummy_state.T()
            Tseal = point.seal_gas_temperature
            if Tseal == None:
                Tseal = Q_(0, "kelvin")
            Ts1r = (ms1f * Ts1f + mend * Tend + 0.95 * mseal * Tseal) / (
                ms1f + mend + 0.95 * mseal
            )
            point.Ts1r = Ts1r
            k_end = k_seal(flow_m=mend, state_up=point.disch, state_down=point.suc)
            self.k_end_seal.append(k_end)
            suc_rotor = State(p=point.suc.p(), T=point.Ts1r, fluid=point.suc.fluid)

            # calculate power losses
            if bearing_mechanical_losses:
                power_losses = (
                    point.oil_density_de
                    * point.oil_flow_journal_bearing_de
                    * point.oil_specific_heat_de
                    * (point.oil_outlet_temperature_de - point.oil_inlet_temperature)
                    + point.oil_density_nde
                    * point.oil_flow_journal_bearing_nde
                    * point.oil_specific_heat_nde
                    * (point.oil_outlet_temperature_nde - point.oil_inlet_temperature)
                    + point.oil_density_nde
                    * point.oil_flow_thrust_bearing_nde
                    * point.oil_specific_heat_nde
                    * (point.oil_outlet_temperature_nde - point.oil_inlet_temperature)
                )
            else:
                power_losses = 0

            test_points_rotor.append(
                Point(
                    suc=suc_rotor,
                    disch=point.disch,
                    flow_m=ms1r,
                    speed=point.speed,
                    b=point.b,
                    D=point.D,
                    power_losses=power_losses,
                    surface_roughness=point.surface_roughness,
                    casing_area=point.casing_area,
                    casing_temperature=point.casing_temperature,
                    ambient_temperature=point.ambient_temperature,
                    convection_constant=point.convection_constant,
                )
            )

        self.points_rotor_t = test_points_rotor

        # convert points_rotor_t to points_rotor_sp
        self.points_rotor_sp = []
        self.points_flange_sp = []
        # calculate ms1r for the guarantee point
        for point, k in zip(self.points_rotor_t, self.k_end_seal):
            error = 1
            # initial estimate of Ts1r_sp with Ts1f_sp
            Ts1r_sp = guarantee_point.suc.T()
            initial_suc = copy(guarantee_point.suc)
            i = 0
            while error > 1e-5 and i < 5:
                initial_suc.update(p=initial_suc.p(), T=Ts1r_sp)
                initial_point_rotor_sp = Point.convert_from(
                    original_point=point,
                    suc=initial_suc,
                    speed=self.speed_operational,
                    find="volume_ratio",
                    reynolds_correction=self.reynolds_correction,
                )
                ms1r_sp = initial_point_rotor_sp.flow_m
                mend_sp = flow_m_seal(
                    k,
                    state_up=initial_point_rotor_sp.disch,
                    state_down=initial_point_rotor_sp.suc,
                )
                ms1f_sp = ms1r_sp - mend_sp
                Ts1f_sp = guarantee_point.suc.T()
                # dummy state to calculate Tend
                dummy_state = copy(initial_point_rotor_sp.disch)
                dummy_state.update(p=initial_point_rotor_sp.suc.p(), h=dummy_state.h())
                Tend_sp = dummy_state.T()
                Ts1r_sp_new = (ms1f_sp * Ts1f_sp + mend_sp * Tend_sp) / (
                    ms1f_sp + mend_sp
                )
                i += 1
                error = abs(Ts1r_sp_new.m - Ts1r_sp.m)
                Ts1r_sp = Ts1r_sp_new
            self.points_rotor_sp.append(initial_point_rotor_sp)
            self.points_flange_sp.append(
                Point(
                    suc=guarantee_point.suc,
                    disch=initial_point_rotor_sp.disch,
                    flow_m=ms1f_sp,
                    speed=speed_operational,
                    b=guarantee_point.b,
                    D=guarantee_point.D,
                    power_losses=initial_point_rotor_sp.power_losses,
                    surface_roughness=guarantee_point.surface_roughness,
                    casing_area=guarantee_point.casing_area,
                    casing_temperature=guarantee_point.casing_temperature,
                    ambient_temperature=guarantee_point.ambient_temperature,
                    convection_constant=guarantee_point.convection_constant,
                )
            )

        super().__init__(self.points_flange_sp)

    def _dict_to_save(self):
        dict_to_save = {
            "reynolds_correction": self.reynolds_correction,
            "speed_operational": str(self.speed_operational),
        }
        # add points to file
        dict_to_save["guarantee_point"] = self.guarantee_point._dict_to_save()

        dict_to_save["test_points"] = {
            f"Point{i}": point._dict_to_save()
            for i, point in enumerate(self.test_points)
        }
        return dict_to_save

    def save(self, file):
        """Save compressor as .toml file.

        Parameters
        ----------
        file : str
            File name.
        """
        with open(file, mode="w") as f:
            toml.dump(self._dict_to_save(), f)

    @classmethod
    def load(cls, file):
        """Load compressor from .toml file.

        Parameters
        ----------
        file : str
            File name.
        """
        parameters = toml.load(file)
        kwargs = {"speed_operational": Q_(parameters.pop("speed_operational", None))}
        # guarantee_point, test_points, speed=None, reynolds_correction=False

        for k, v in parameters.items():
            if "guarantee_point" in k:
                kwargs[k] = Point(**Point._dict_from_load(v))
            elif "test_points" in k:
                kwargs[k] = [Point1Sec(**Point._dict_from_load(v)) for v in v.values()]
            else:
                kwargs[k] = v

        return cls(**kwargs)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if (
                self.reynolds_correction == other.reynolds_correction
                and self.speed_operational == other.speed_operational
                and self.guarantee_point == other.guarantee_point
            ):
                test_points_other = sorted(other.test_points, key=lambda x: x.flow_v)
                test_points_self = sorted(self.test_points, key=lambda x: x.flow_v)
                if len(test_points_self) == len(test_points_other):
                    if test_points_self == test_points_other:
                        return True

    def calculate_speed_to_match_discharge_pressure(self):
        """Calculate the speed to match the discharge pressure of the guarantee point."""

        def calculate_disch_pressure_delta(x):
            compressor = StraightThrough(
                guarantee_point=self.guarantee_point,
                test_points=self.test_points,
                speed_operational=x,
                reynolds_correction=self.reynolds_correction,
                bearing_mechanical_losses=self.bearing_mechanical_losses,
            )

            point = compressor.point(flow_m=self.guarantee_point.flow_m, speed=x)
            # add 1 pascal to guarantee that discharge pressure is higher
            return point.disch.p().m - (self.guarantee_point.disch.p().m + 1)

        new_speed = newton(calculate_disch_pressure_delta, self.speed_operational.m)
        return self.__class__(
            guarantee_point=self.guarantee_point,
            test_points=self.test_points,
            speed_operational=new_speed,
            reynolds_correction=self.reynolds_correction,
            bearing_mechanical_losses=self.bearing_mechanical_losses,
        )


class PointFirstSection(Point):
    """Point class for a compressor with 2 sections in a back-to-back configuration.

    Parameters
    ----------
    balance_line_flow_m : float, pint.Quantity
        Balance line mass flow (kg/s).
    seal_gas_flow_m : float, pint.Quantity
        Seal gas mass flow (kg/s).
    seal_gas_temperature : float, pint.Quantity
        Seal gas injection temperature (degK).
    first_section_discharge_flow_m : float, pint.Quantity
        First section discharge mass flow (kg/s).
    div_wall_flow_m : float, pint.Quantity
        Division wall mass flow (kg/s).
    end_seal_upstream_temperature : float, pint.Quantity
        Temperature upstream the end seal (degK).
    end_seal_upstream_pressure : float, pint.Quantity
        Pressure upstream the end seal (Pa).
    div_wall_upstream_temperature : float, pint.Quantity
        Temperature upstream the division wall (degK).
    div_wall_upstream_pressure : float, pint.Quantity
         Pressure upstream the division wall (Pa).
    oil_flow_journal_bearing_de : float, pint.Quantity
        Oil flow journal bearing drive end side (m³/s).
    oil_flow_journal_bearing_nde : float, pint.Quantity
        Oil flow journal bearing non-drive end side (m³/s).
    oil_flow_thrust_bearing_nde : float, pint.Quantity
         Oil flow journal bearing thrust bearing (m³/s).
    oil_inlet_temperature : float, pint.Quantity
        Oil inlet temperature (degK).
    oil_outlet_temperature_de : float, pint.Quantity
        Oil outlet temperature journal bearing drive end side (degK).
    oil_outlet_temperature_nde : float, pint.Quantity
        Oil outlet temperature bearing non-drive end side (degK).
    oil_specific_heat_de : float, pint.Quantity
        Oil specific heat bearing drive end side (J/kg/degK).
    oil_specific_heat_nde : float, pint.Quantity
        Oil specific heat bearing non-drive end side (J/kg/degK).
    oil_density_de : float, pint.Quantity
        Oil Density bearing drive end side (kg/m³).
    oil_density_nde : float, pint.Quantity
        Oil density bearing non-drive end side (kg/m³).

    Returns
    -------
    point1sec : ccp.Point2Sec
        A point for a back to back compressor (inherited from ccp.Point).

    Examples
    --------
    >>> import ccp
    >>> p = PointFirstSection(
    ...     flow_m=Q_(4.325, "kg/s"),
    ...     speed=Q_(9096, "RPM"),
    ...     b=Q_(10.5, "mm"),
    ...     D=Q_(365, "mm"),
    ...     suc=State(
    ...         p=Q_(5.182, "bar"),
    ...         T=Q_(299.5, "degK"),
    ...         fluid={
    ...             "carbon dioxide": 1,
    ...         },
    ...     ),
    ...     disch=State(
    ...         p=Q_(14.95, "bar"),
    ...         T=Q_(397.6, "degK"),
    ...         fluid={
    ...             "carbon dioxide": 1,
    ...         },
    ...     ),
    ...     balance_line_flow_m=Q_(0.1625, "kg/s"),
    ...     seal_gas_flow_m=Q_(0.0616, "kg/s"),
    ...     seal_gas_temperature=299.7,
    ...     first_section_discharge_flow_m=4.8059,
    ...     div_wall_flow_m=None,
    ...     end_seal_upstream_temperature=304.3,
    ...     end_seal_upstream_pressure=14.59,
    ...     div_wall_upstream_temperature=362.5,
    ...     div_wall_upstream_pressure=26.15,
    ...     oil_flow_journal_bearing_de=Q_(31.515, "l/min"),
    ...     oil_flow_journal_bearing_nde=Q_(22.67, "l/min"),
    ...     oil_flow_thrust_bearing_nde=Q_(126.729, "l/min"),
    ...     oil_inlet_temperature=Q_(41.544, "degC"),
    ...     oil_outlet_temperature_de=Q_(49.727, "degC"),
    ...     oil_outlet_temperature_nde=Q_(50.621, "degC"),
    ...     oil_specific_heat_de=Q_(2.02, "kJ/kg/degK"),
    ...     oil_specific_heat_nde=Q_(2.02, "kJ/kg/degK"),
    ...     oil_density_de=Q_(846.9, "kg/m³"),
    ...     oil_density_nde=Q_(846.9, "kg/m³"),
    ...     casing_area=5.5,
    ...     casing_temperature=Q_(23.895, "degC"),
    ...     ambient_temperature=Q_(0, "degC"),
    ... )
    """

    @check_units
    def __init__(
        self,
        *args,
        balance_line_flow_m=None,
        seal_gas_flow_m=None,
        seal_gas_temperature=None,
        first_section_discharge_flow_m=None,
        div_wall_flow_m=None,
        end_seal_upstream_temperature=None,
        end_seal_upstream_pressure=None,
        div_wall_upstream_temperature=None,
        div_wall_upstream_pressure=None,
        oil_flow_journal_bearing_de=None,
        oil_flow_journal_bearing_nde=None,
        oil_flow_thrust_bearing_nde=None,
        oil_inlet_temperature=None,
        oil_outlet_temperature_de=None,
        oil_outlet_temperature_nde=None,
        oil_specific_heat_de=None,
        oil_specific_heat_nde=None,
        oil_density_de=None,
        oil_density_nde=None,
        leakages=True,
        bearing_mechanical_losses=True,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )
        self._args = args
        self._kwargs = kwargs
        self.seal_gas_flow_m = seal_gas_flow_m
        self.seal_gas_temperature = seal_gas_temperature

        # keep the following original values used in instantiation
        self._balance_line_flow_m = balance_line_flow_m
        self._div_wall_flow_m = div_wall_flow_m
        self._first_section_discharge_flow_m = first_section_discharge_flow_m

        self.balance_line_flow_m = balance_line_flow_m
        self.first_section_discharge_flow_m = first_section_discharge_flow_m
        self.div_wall_flow_m = div_wall_flow_m
        self.end_seal_upstream_temperature = end_seal_upstream_temperature
        self.end_seal_upstream_pressure = end_seal_upstream_pressure
        self.div_wall_upstream_temperature = div_wall_upstream_temperature
        self.div_wall_upstream_pressure = div_wall_upstream_pressure
        self.oil_flow_journal_bearing_de = oil_flow_journal_bearing_de
        self.oil_flow_journal_bearing_nde = oil_flow_journal_bearing_nde
        self.oil_flow_thrust_bearing_nde = oil_flow_thrust_bearing_nde
        self.oil_inlet_temperature = oil_inlet_temperature
        self.oil_outlet_temperature_de = oil_outlet_temperature_de
        self.oil_outlet_temperature_nde = oil_outlet_temperature_nde
        self.oil_specific_heat_de = oil_specific_heat_de
        self.oil_specific_heat_nde = oil_specific_heat_nde
        self.oil_density_de = oil_density_de
        self.oil_density_nde = oil_density_nde
        self.leakages = leakages
        self.bearing_mechanical_losses = bearing_mechanical_losses

        # check case for no leakage. If no leakage, div_wall_flow_m and first_section_discharge_flow_m are zero
        # and we equate the seal/div wall states to avoid errors
        if not self.leakages:
            self.div_wall_flow_m = Q_(0, "kg/s")
            self.balance_line_flow_m = Q_(0, "kg/s")
            self.seal_gas_flow_m = Q_(0, "kg/s")
            self.seal_gas_temperature = self.suc.T()
            self.end_seal_upstream_temperature = self.suc.T()
            self.end_seal_upstream_pressure = self.suc.p()
            self.div_wall_upstream_temperature = self.disch.T()
            self.div_wall_upstream_pressure = self.disch.p()

        if not self.bearing_mechanical_losses:
            self.oil_flow_journal_bearing_de = Q_(0, "m³/s")
            self.oil_flow_journal_bearing_nde = Q_(0, "m³/s")
            self.oil_flow_thrust_bearing_nde = Q_(0, "m³/s")
            self.oil_inlet_temperature = self.suc.T()
            self.oil_outlet_temperature_de = self.suc.T()
            self.oil_outlet_temperature_nde = self.suc.T()
            oil_specific_heat_de = (Q_(2.02, "kJ/kg/degK"),)
            oil_specific_heat_nde = (Q_(2.02, "kJ/kg/degK"),)
            oil_density_de = (Q_(846.9, "kg/m³"),)
            oil_density_nde = (Q_(846.9, "kg/m³"),)

        self.end_seal_upstream_state = State(
            p=self.end_seal_upstream_pressure,
            T=self.end_seal_upstream_temperature,
            fluid=self.suc.fluid,
        )
        self.end_seal_downstream_state = copy(self.suc)
        self.end_seal_downstream_state.update(
            p=self.end_seal_downstream_state.p(), h=self.end_seal_upstream_state.h()
        )

        self.div_wall_upstream_state = State(
            p=self.div_wall_upstream_pressure,
            T=self.div_wall_upstream_temperature,
            fluid=self.suc.fluid,
        )
        self.div_wall_downstream_state = copy(self.disch)
        self.div_wall_downstream_state.update(
            p=self.div_wall_downstream_state.p(), h=self.div_wall_upstream_state.h()
        )

    def _dict_to_save(self):
        """Returns a dict that will be saved to a toml file."""
        dict_to_save = super()._dict_to_save()
        parameters = [
            "balance_line_flow_m",
            "first_section_discharge_flow_m",
            "div_wall_flow_m",
        ]
        for parameter in parameters:
            if getattr(self, "_" + parameter) is not None:
                dict_to_save[parameter] = str(getattr(self, "_" + parameter))

        parameters = [
            "seal_gas_flow_m",
            "seal_gas_temperature",
            "end_seal_upstream_temperature",
            "end_seal_upstream_pressure",
            "div_wall_upstream_temperature",
            "div_wall_upstream_pressure",
            "oil_flow_journal_bearing_de",
            "oil_flow_journal_bearing_nde",
            "oil_flow_thrust_bearing_nde",
            "oil_inlet_temperature",
            "oil_outlet_temperature_de",
            "oil_outlet_temperature_nde",
            "oil_specific_heat_de",
            "oil_specific_heat_nde",
            "oil_density_de",
            "oil_density_nde",
        ]
        for parameter in parameters:
            if getattr(self, parameter) is not None:
                dict_to_save[parameter] = str(getattr(self, parameter))

        return dict_to_save


class PointSecondSection(Point1Sec):
    """Point for second section"""

    pass


class BackToBack(Impeller):
    """Back to Back compressor"""

    @check_units
    def __init__(
        self,
        guarantee_point_sec1,
        test_points_sec1,
        guarantee_point_sec2,
        test_points_sec2,
        reynolds_correction=False,
        bearing_mechanical_losses=False,
        speed_operational=None,
    ):
        self.guarantee_point_sec1 = guarantee_point_sec1
        self.guarantee_point_sec2 = guarantee_point_sec2
        self.test_points_sec1 = test_points_sec1
        self.test_points_sec2 = test_points_sec2
        if speed_operational is None:
            speed_operational = guarantee_point_sec1.speed
        self.speed_operational = speed_operational
        self.reynolds_correction = reynolds_correction
        self.bearing_mechanical_losses = bearing_mechanical_losses

        # points for test flange conditions
        self.points_flange_t_sec1 = test_points_sec1
        self.points_flange_t_sec2 = test_points_sec2

        # calculate rotor condition for sec1
        test_points_sec1_rotor = np.full(len(test_points_sec1), np.nan, dtype=object)
        self.k_end_seal = np.zeros(
            len(test_points_sec1), dtype=object
        )  # array with seal constants
        self.k_div_wall = np.zeros(
            len(test_points_sec1), dtype=object
        )  # array with div wall seal constants

        for point in test_points_sec1:
            if point.div_wall_flow_m:
                point.first_section_discharge_flow_m = (
                    point.flow_m
                    + point.div_wall_flow_m
                    + point.balance_line_flow_m
                    + 0.95 * point.seal_gas_flow_m / 2
                )
        for i, point in enumerate(test_points_sec1):
            # calculate power losses
            if bearing_mechanical_losses:
                power_losses = (
                    point.oil_density_de
                    * point.oil_flow_journal_bearing_de
                    * point.oil_specific_heat_de
                    * (point.oil_outlet_temperature_de - point.oil_inlet_temperature)
                    + point.oil_density_nde
                    * point.oil_flow_journal_bearing_nde
                    * point.oil_specific_heat_nde
                    * (point.oil_outlet_temperature_nde - point.oil_inlet_temperature)
                    + point.oil_density_nde
                    * point.oil_flow_thrust_bearing_nde
                    * point.oil_specific_heat_nde
                    * (point.oil_outlet_temperature_nde - point.oil_inlet_temperature)
                )
            else:
                power_losses = 0
            # Here we check for _first_section_discharge_flow_m because we want to use the
            # value given in the test point, not the calculated value.
            # This way we can guarantee that everytime we create the compressor, the same
            # values are used, and we also get the same results.
            if point._first_section_discharge_flow_m:
                md1f_t = point._first_section_discharge_flow_m
                ms1f_t = point.flow_m
                mbal_t = point._balance_line_flow_m
                mseal_t = point.seal_gas_flow_m
                mend_t = mbal_t - (0.95 * mseal_t) / 2
                point.end_seal_flow_m = mend_t
                mdiv_t = md1f_t - ms1f_t - mend_t - 0.95 * mseal_t
                point.div_wall_flow_m = mdiv_t
                ms1r_t = ms1f_t + mend_t + 0.95 * mseal_t

                ps1f_t = point.suc.p()
                Ts1f_t = point.suc.T()

                pd1f_t = point.disch.p()
                Td1f_t = point.disch.T()
                pd2f_t = point.div_wall_upstream_pressure
                Td2f_t = point.div_wall_upstream_temperature
                Tseal_t = point.seal_gas_temperature

                Tend_t = point.end_seal_downstream_state.T()

                k_end = k_seal(
                    flow_m=mend_t,
                    state_up=point.end_seal_upstream_state,
                    state_down=point.end_seal_downstream_state,
                )
                self.k_end_seal[i] = k_end

                k_div_wall = k_seal(
                    flow_m=mdiv_t,
                    state_up=point.div_wall_upstream_state,
                    state_down=point.div_wall_downstream_state,
                )
                self.k_div_wall[i] = k_div_wall
                Tdiv_t = point.div_wall_downstream_state.T()
                Ts1r_t = (
                    ms1f_t * Ts1f_t + mend_t * Tend_t + 0.95 * mseal_t * Tseal_t
                ) / (ms1f_t + mend_t + 0.95 * mseal_t)
                Td1r_t = (md1f_t * Td1f_t - mdiv_t * Tdiv_t) / (md1f_t - mdiv_t)
                suc_sec1_rotor_t = State(p=ps1f_t, T=Ts1r_t, fluid=point.suc.fluid)
                disch_sec1_rotor_t = State(p=pd1f_t, T=Td1r_t, fluid=point.suc.fluid)

                test_points_sec1_rotor[i] = Point(
                    suc=suc_sec1_rotor_t,
                    disch=disch_sec1_rotor_t,
                    flow_m=ms1r_t,
                    speed=point.speed,
                    b=point.b,
                    D=point.D,
                    power_losses=power_losses,
                    casing_area=point.casing_area,
                    casing_temperature=point.casing_temperature,
                    ambient_temperature=point.ambient_temperature,
                    surface_roughness=point.surface_roughness,
                    convection_constant=point.convection_constant,
                )

            # use calculated k_end_seal and k_div_wall to calculate seal mass flow
            # use manual mean, since np.mean removes the units
            if (self.k_end_seal == 0).all():
                k_end_seal_mean = Q_(
                    0, "kelvin**0.5 kilogram**0.5 mole**0.5/(pascal second)"
                )
            else:
                k_end_seal_mean = sum(self.k_end_seal[self.k_end_seal != 0]) / len(
                    self.k_end_seal[self.k_end_seal != 0]
                )
            if (self.k_div_wall == 0).all():
                k_div_wall_mean = Q_(
                    0, "kelvin**0.5 kilogram**0.5 mole**0.5/(pascal second)"
                )
            else:
                k_div_wall_mean = sum(self.k_div_wall[self.k_div_wall != 0]) / len(
                    self.k_div_wall[self.k_div_wall != 0]
                )

        for i, point in enumerate(test_points_sec1):
            # calculate power losses
            if bearing_mechanical_losses:
                power_losses = (
                    point.oil_density_de
                    * point.oil_flow_journal_bearing_de
                    * point.oil_specific_heat_de
                    * (point.oil_outlet_temperature_de - point.oil_inlet_temperature)
                    + point.oil_density_nde
                    * point.oil_flow_journal_bearing_nde
                    * point.oil_specific_heat_nde
                    * (point.oil_outlet_temperature_nde - point.oil_inlet_temperature)
                    + point.oil_density_nde
                    * point.oil_flow_thrust_bearing_nde
                    * point.oil_specific_heat_nde
                    * (point.oil_outlet_temperature_nde - point.oil_inlet_temperature)
                )
            else:
                power_losses = 0
            if not point._first_section_discharge_flow_m:
                self.k_end_seal[i] = k_end_seal_mean
                end_seal_flow_m = flow_m_seal(
                    k_seal=k_end_seal_mean,
                    state_up=point.end_seal_upstream_state,
                    state_down=point.end_seal_downstream_state,
                )

                self.k_div_wall[i] = k_div_wall_mean
                div_wall_flow_m = flow_m_seal(
                    k_seal=k_div_wall_mean,
                    state_up=point.div_wall_upstream_state,
                    state_down=point.div_wall_downstream_state,
                )

                # calculate balance line and rotor suction flow
                if point.seal_gas_flow_m == None:  ##################
                    point.seal_gas_flow_m = 0
                mseal_t = point.seal_gas_flow_m
                mend_t = end_seal_flow_m
                point.end_seal_flow_m = mend_t

                mbal_t = mend_t + (0.95 * mseal_t) / 2
                point.balance_line_flow_m = mbal_t

                ms1f_t = point.flow_m
                ms1r_t = ms1f_t + mend_t + 0.95 * mseal_t

                # calculate discharge flow
                mdiv_t = div_wall_flow_m
                point.div_wall_flow_m = mdiv_t
                md1f_t = mdiv_t + ms1f_t + mend_t + 0.95 * mseal_t
                point.first_section_discharge_flow_m = md1f_t

                Ts1f_t = point.suc.T()
                Tend_t = point.end_seal_downstream_state.T()
                Tdiv_t = point.div_wall_downstream_state.T()
                Tseal_t = point.seal_gas_temperature
                Td1f_t = point.disch.T()
                ps1f_t = point.suc.p()
                pd1f_t = point.disch.p()

                Ts1r_t = (
                    ms1f_t * Ts1f_t + mend_t * Tend_t + 0.95 * mseal_t * Tseal_t
                ) / (ms1f_t + mend_t + 0.95 * mseal_t)
                Td1r_t = (md1f_t * Td1f_t - mdiv_t * Tdiv_t) / (md1f_t - mdiv_t)

                suc_sec1_rotor_t = State(p=ps1f_t, T=Ts1r_t, fluid=point.suc.fluid)
                disch_sec1_rotor_t = State(p=pd1f_t, T=Td1r_t, fluid=point.suc.fluid)

                test_points_sec1_rotor[i] = Point(
                    suc=suc_sec1_rotor_t,
                    disch=disch_sec1_rotor_t,
                    flow_m=ms1r_t,
                    speed=point.speed,
                    b=point.b,
                    D=point.D,
                    power_losses=power_losses,
                    casing_area=point.casing_area,
                    casing_temperature=point.casing_temperature,
                    ambient_temperature=point.ambient_temperature,
                    surface_roughness=point.surface_roughness,
                    convection_constant=point.convection_constant,
                )

        self.k_end_seal_mean = k_end_seal_mean
        self.k_div_wall_mean = k_div_wall_mean
        self.points_rotor_t_sec1 = test_points_sec1_rotor

        # calculate rotor condition for sec2
        test_points_sec2_rotor = np.full(len(test_points_sec1), np.nan, dtype=object)
        for i, point_f in enumerate(test_points_sec2):
            # calculate power losses
            if bearing_mechanical_losses:
                power_losses = (
                    point.oil_density_de
                    * point.oil_flow_journal_bearing_de
                    * point.oil_specific_heat_de
                    * (point.oil_outlet_temperature_de - point.oil_inlet_temperature)
                    + point.oil_density_nde
                    * point.oil_flow_journal_bearing_nde
                    * point.oil_specific_heat_nde
                    * (point.oil_outlet_temperature_nde - point.oil_inlet_temperature)
                    + point.oil_density_nde
                    * point.oil_flow_thrust_bearing_nde
                    * point.oil_specific_heat_nde
                    * (point.oil_outlet_temperature_nde - point.oil_inlet_temperature)
                )
            else:
                power_losses = 0

            ms2f_t = point_f.flow_m
            mbal_t = point_f.balance_line_flow_m
            if point_f.seal_gas_flow_m == None:  ##################
                point_f.seal_gas_flow_m = 0
            mseal_t = point_f.seal_gas_flow_m
            mend_t = mbal_t - 0.95 * mseal_t / 2
            ms2r_t = ms2f_t - mend_t
            point_r = Point(
                suc=point_f.suc,
                disch=point_f.disch,
                flow_m=ms2r_t,
                speed=point_f.speed,
                b=point_f.b,
                D=point_f.D,
                power_losses=power_losses,
                casing_area=point_f.casing_area,
                casing_temperature=point_f.casing_temperature,
                ambient_temperature=point_f.ambient_temperature,
                surface_roughness=point_f.surface_roughness,
                convection_constant=point_f.convection_constant,
            )
            test_points_sec2_rotor[i] = point_r

        self.points_rotor_t_sec2 = test_points_sec2_rotor

        # convert sec2 points_rotor_sp to points_flange_sp
        self.points_rotor_sp_sec2 = []
        self.points_flange_sp_sec2 = []

        # calculate rotor specified conditions for sec1
        self.points_rotor_sp_sec1 = []
        ms1f_sp_array = np.zeros(len(self.points_rotor_t_sec1), dtype=object)
        for i, point, k_end_seal, k_div_wall in zip(
            range(len(ms1f_sp_array)),
            self.points_rotor_t_sec1,
            self.k_end_seal,
            self.k_div_wall,
        ):
            initial_point = Point.convert_from(
                original_point=point,
                suc=guarantee_point_sec1.suc,
                speed=self.speed_operational,
                find="volume_ratio",
                reynolds_correction=self.reynolds_correction,
            )
            # determine rotor specified suction state
            end_seal_state_upstream_sp = State(
                p=initial_point.disch.p(),
                T=guarantee_point_sec2.suc.T(),
                fluid=initial_point.suc.fluid,
            )
            end_seal_state_downstream_sp = copy(initial_point.disch)
            end_seal_state_downstream_sp.update(
                p=guarantee_point_sec1.suc.p(), h=end_seal_state_upstream_sp.h()
            )
            Tend_sp = end_seal_state_downstream_sp.T()

            mend_sp = flow_m_seal(
                k_seal=k_end_seal,
                state_up=end_seal_state_upstream_sp,
                state_down=end_seal_state_downstream_sp,
            )

            Ts1f_sp = guarantee_point_sec1.suc.T()
            qs1r_sp = flow_from_phi(
                D=point.D, phi=point.phi, speed=self.speed_operational
            )
            ps1r_sp = guarantee_point_sec1.suc.p()
            vs1f_sp = guarantee_point_sec1.suc.v()
            dummy_suc = copy(guarantee_point_sec1.suc)

            error = 1
            dm = Q_(1, "kg/s")
            ms1f_sp = qs1r_sp / vs1f_sp  # initial guess
            while error > 0.00001:
                ms1r_sp = ms1f_sp + mend_sp
                Ts1r_sp = (mend_sp * Tend_sp + ms1f_sp * Ts1f_sp) / ms1r_sp
                dummy_suc.update(p=ps1r_sp, T=Ts1r_sp)
                vs1r_sp = dummy_suc.v()
                qs1r_sp_1 = ms1r_sp * vs1r_sp

                fx = -qs1r_sp + qs1r_sp_1
                ms1f_sp_new = ms1f_sp + dm
                ms1r_sp_new = ms1f_sp_new + mend_sp
                Ts1r_sp_new = (ms1f_sp_new * Ts1f_sp + mend_sp * Tend_sp) / ms1r_sp_new
                dummy_suc.update(p=ps1r_sp, T=Ts1r_sp_new)
                vs1r_sp_new = dummy_suc.v()
                qs1r_sp_1_new = ms1r_sp_new * vs1r_sp_new
                dfx = (qs1r_sp_1_new - qs1r_sp_1) / dm
                ms1f_sp = ms1f_sp - (fx / dfx)
                error = ((fx**2) ** 0.5).m

            ms1f_sp_array[i] = ms1f_sp

            rotor_sp_sec1_suc = ccp.State(
                p=guarantee_point_sec1.suc.p(),
                T=Ts1r_sp,
                fluid=guarantee_point_sec1.suc.fluid,
            )
            point_r_sp = Point.convert_from(
                original_point=point,
                suc=rotor_sp_sec1_suc,
                speed=self.speed_operational,
                find="volume_ratio",
                reynolds_correction=self.reynolds_correction,
            )
            self.points_rotor_sp_sec1.append(point_r_sp)
        self.imp_rotor_sp_sec1 = Impeller(self.points_rotor_sp_sec1)

        # estimate rotor guarantee flow using datasheet conditions
        end_seal_flow_m_sp = flow_m_seal(
            k_seal=k_end_seal,
            state_up=guarantee_point_sec2.suc,
            state_down=guarantee_point_sec1.suc,
        )
        # calculate suction pressure for sec2
        ms1r_sp = guarantee_point_sec1.flow_m + end_seal_flow_m_sp

        disch_r_sp = self.imp_rotor_sp_sec1.point(
            flow_m=ms1r_sp, speed=self.speed_operational
        ).disch
        ps2f_sp = disch_r_sp.p() - (
            guarantee_point_sec1.disch.p() - guarantee_point_sec2.suc.p()
        )
        suc2f_sp = State(
            p=ps2f_sp,
            T=guarantee_point_sec2.suc.T(),
            fluid=guarantee_point_sec2.suc.fluid,
        )
        for point_r_t in self.points_rotor_t_sec2:
            point_r_sp = Point.convert_from(
                original_point=point_r_t,
                suc=suc2f_sp,
                speed=self.speed_operational,
                find="volume_ratio",
                reynolds_correction=self.reynolds_correction,
            )
            self.points_rotor_sp_sec2.append(point_r_sp)
            mend_sp = flow_m_seal(
                k_seal=k_end_seal,
                state_up=guarantee_point_sec2.suc,
                state_down=guarantee_point_sec1.suc,
            )
            ms2r_sp = point_r_sp.flow_m
            ms2f_sp = ms2r_sp + mend_sp
            point_sp = Point(
                suc=point_r_sp.suc,
                disch=point_r_sp.disch,
                flow_m=ms2f_sp,
                speed=point_r_sp.speed,
                power_losses=point_r_sp.power_losses,
                b=point_r_sp.b,
                D=point_r_sp.D,
            )
            self.points_flange_sp_sec2.append(point_sp)
        # change points_flange power to real power calculated in rotor point
        for p_flange, p_rotor in zip(
            self.points_flange_sp_sec2, self.points_rotor_sp_sec2
        ):
            p_flange.power = p_rotor.power

        self.imp_flange_sp_sec2 = Impeller(self.points_flange_sp_sec2)

        # convert sec1 points_rotor_t to points_rotor_sp
        self.points_flange_sp_sec1 = []
        for point_r_sp, ms1f_sp, k_end_seal, k_div_wall in zip(
            self.points_rotor_sp_sec1, ms1f_sp_array, self.k_end_seal, self.k_div_wall
        ):
            # calculate div wall flow
            suc_sec2 = ccp.State(
                p=point_r_sp.disch.p(),
                T=guarantee_point_sec2.suc.T(),
                fluid=guarantee_point_sec2.suc.fluid,
            )
            imp_sec2_conv = self.imp_flange_sp_sec2
            imp_sec2_flow_m = point_r_sp.flow_m
            if imp_sec2_flow_m > imp_sec2_conv.flow_m.max():
                imp_sec2_flow_m = imp_sec2_conv.flow_m.max()
            sec2_point = imp_sec2_conv.point(
                flow_m=imp_sec2_flow_m, speed=self.speed_operational
            )
            sec2_disch = ccp.point.disch_from_suc_head_eff(
                suc=suc_sec2, head=sec2_point.head, eff=sec2_point.eff
            )
            mdiv_sp = flow_m_seal(
                k_seal=k_div_wall,
                state_up=sec2_disch,
                state_down=point_r_sp.disch,
            )
            div_wall_downstream_state = copy(sec2_disch)
            div_wall_downstream_state.update(
                p=point_r_sp.disch.p(), h=div_wall_downstream_state.h()
            )
            Tdiv_sp = div_wall_downstream_state.T()

            # calculate flange disch
            Td1r_sp = point_r_sp.disch.T()
            Td1f_sp = (ms1r_sp * Td1r_sp + mdiv_sp * Tdiv_sp) / (ms1r_sp + mdiv_sp)
            disch_f = ccp.State(
                p=point_r_sp.disch.p(), T=Td1f_sp, fluid=guarantee_point_sec1.suc.fluid
            )
            point_flange = Point(
                suc=guarantee_point_sec1.suc,
                disch=disch_f,
                flow_m=ms1f_sp,
                speed=self.speed_operational,
                b=guarantee_point_sec1.b,
                D=guarantee_point_sec1.D,
            )
            self.points_flange_sp_sec1.append(point_flange)
            self.points_flange_sp_sec1[-1].div_wall_flow_m = mdiv_sp
            self.points_flange_sp_sec1[-1].first_section_discharge_flow_m = (
                mdiv_sp + ms1r_sp
            )
            # change points_flange power to real power calculated in rotor point
            for p_flange, p_rotor in zip(
                self.points_flange_sp_sec1, self.points_rotor_sp_sec1
            ):
                p_flange.power = p_rotor.power
                p_flange.power_losses = p_rotor.power_losses
        self.imp_flange_sp_sec1 = Impeller(self.points_flange_sp_sec1)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if (
                self.reynolds_correction == other.reynolds_correction
                and self.speed_operational == other.speed_operational
                and self.guarantee_point_sec1 == other.guarantee_point_sec1
                and self.guarantee_point_sec2 == other.guarantee_point_sec2
            ):
                test_points_sec1_other = sorted(
                    other.test_points_sec1, key=lambda x: x.flow_v
                )
                test_points_sec1_self = sorted(
                    self.test_points_sec1, key=lambda x: x.flow_v
                )
                test_points_sec2_other = sorted(
                    other.test_points_sec2, key=lambda x: x.flow_v
                )
                test_points_sec2_self = sorted(
                    self.test_points_sec2, key=lambda x: x.flow_v
                )
                if len(test_points_sec1_self) == len(test_points_sec1_other) and len(
                    test_points_sec2_self
                ) == len(test_points_sec2_other):
                    if (
                        test_points_sec1_self == test_points_sec1_other
                        and test_points_sec2_self == test_points_sec2_other
                    ):
                        return True

    def _dict_to_save(self):
        dict_to_save = {
            "reynolds_correction": self.reynolds_correction,
            "speed_operational": str(self.speed_operational),
        }
        # add points to file
        dict_to_save["guarantee_point_sec1"] = self.guarantee_point_sec1._dict_to_save()
        dict_to_save["guarantee_point_sec2"] = self.guarantee_point_sec2._dict_to_save()
        dict_to_save["test_points_sec1"] = {
            f"Point{i}": point._dict_to_save()
            for i, point in enumerate(self.test_points_sec1)
        }
        dict_to_save["test_points_sec2"] = {
            f"Point{i}": point._dict_to_save()
            for i, point in enumerate(self.test_points_sec2)
        }
        return dict_to_save

    def save(self, file):
        """Save compressor as .toml file.

        Parameters
        ----------
        file : str
            File name.
        """
        with open(file, mode="w") as f:
            toml.dump(self._dict_to_save(), f)

    @classmethod
    def load(cls, file):
        """Load compressor from .toml file.

        Parameters
        ----------
        file : str
            File name.
        """
        parameters = toml.load(file)
        kwargs = {"speed_operational": Q_(parameters.pop("speed_operational", None))}

        for k, v in parameters.items():
            if "guarantee_point" in k:
                kwargs[k] = Point(**Point._dict_from_load(v))
            elif "test_points_sec1" in k:
                kwargs[k] = [
                    PointFirstSection(**Point._dict_from_load(v)) for v in v.values()
                ]
            elif "test_points_sec2" in k:
                kwargs[k] = [
                    PointSecondSection(**Point._dict_from_load(v)) for v in v.values()
                ]
            else:
                kwargs[k] = v

        return cls(**kwargs)

    def point_sec1(self, *args, **kwargs):
        # calculate flange point from impeller object
        p_sec1 = self.imp_flange_sp_sec1.point(*args, **kwargs)

        # calculate rotor flow
        ps2f_sp = p_sec1.disch.p() - (
            self.guarantee_point_sec1.disch.p() - self.guarantee_point_sec2.suc.p()
        )
        suc2f_sp = State(
            p=ps2f_sp,
            T=self.guarantee_point_sec2.suc.T(),
            fluid=self.guarantee_point_sec2.suc.fluid,
        )
        end_seal_flow_m_sp = flow_m_seal(
            k_seal=self.k_end_seal_mean,
            state_up=suc2f_sp,
            state_down=p_sec1.suc,
        )
        ms1r_sp = p_sec1.flow_m + end_seal_flow_m_sp

        # calculate 'real' power from rotor conditions
        p_sec1_rotor = self.imp_rotor_sp_sec1.point(flow_m=ms1r_sp, speed=p_sec1.speed)
        p_sec1.power = p_sec1_rotor.power
        p_sec1.power_losses = p_sec1_rotor.power_losses
        p_sec1.power_shaft = p_sec1_rotor.power_shaft

        # set mach number and mach_diff
        mach_interpolation = parameter_interpolation(
            p_sec1.phi,
            [p.phi for p in self.points_flange_t_sec1],
            [p.mach for p in self.points_flange_t_sec1],
        )
        p_sec1.mach = mach_interpolation
        p_sec1.mach_diff = mach_interpolation - self.guarantee_point_sec1.mach

        # set reynolds number and reynolds_ratio
        reynolds_interpolation = parameter_interpolation(
            p_sec1.phi,
            [p.phi for p in self.points_flange_t_sec1],
            [p.reynolds for p in self.points_flange_t_sec1],
        )
        p_sec1.reynolds = reynolds_interpolation
        p_sec1.reynolds_ratio = (
            reynolds_interpolation / self.guarantee_point_sec1.reynolds
        )

        # set new volume ratio ratio
        p_sec1.volume_ratio_ratio = (
            p_sec1.volume_ratio / self.guarantee_point_sec1.volume_ratio
        )

        return p_sec1

    def point_sec2(self, *args, **kwargs):
        # calculate flange point from impeller object
        p_sec2 = self.imp_flange_sp_sec2.point(*args, **kwargs)

        # calculate 'real' power from rotor conditions
        end_seal_flow_m_sp = flow_m_seal(
            k_seal=self.k_end_seal_mean,
            state_up=p_sec2.suc,
            state_down=self.guarantee_point_sec1.suc,
        )
        ms2r_sp = p_sec2.flow_m - end_seal_flow_m_sp
        p_sec2.power = ms2r_sp * p_sec2.head / p_sec2.eff
        p_sec2.power_shaft = p_sec2.power + p_sec2.power_losses

        # set mach number and mach_diff
        mach_interpolation = parameter_interpolation(
            p_sec2.phi,
            [p.phi for p in self.points_flange_t_sec2],
            [p.mach for p in self.points_flange_t_sec2],
        )
        p_sec2.mach = mach_interpolation
        p_sec2.mach_diff = mach_interpolation - self.guarantee_point_sec2.mach

        # set reynolds number and reynolds_ratio
        reynolds_interpolation = parameter_interpolation(
            p_sec2.phi,
            [p.phi for p in self.points_flange_t_sec2],
            [p.reynolds for p in self.points_flange_t_sec2],
        )
        p_sec2.reynolds = reynolds_interpolation
        p_sec2.reynolds_ratio = (
            reynolds_interpolation / self.guarantee_point_sec2.reynolds
        )
        # set new volume ratio ratio
        p_sec2.volume_ratio_ratio = (
            p_sec2.volume_ratio / self.guarantee_point_sec2.volume_ratio
        )

        return p_sec2

    def calculate_speed_to_match_discharge_pressure(self):
        """Calculate the speed to match the discharge pressure of the guarantee point."""

        def calculate_disch_pressure_delta(x):
            compressor = BackToBack(
                guarantee_point_sec1=self.guarantee_point_sec1,
                guarantee_point_sec2=self.guarantee_point_sec2,
                test_points_sec1=self.test_points_sec1,
                test_points_sec2=self.test_points_sec2,
                speed_operational=x,
                reynolds_correction=self.reynolds_correction,
                bearing_mechanical_losses=self.bearing_mechanical_losses,
            )

            point = compressor.point_sec2(
                flow_m=self.guarantee_point_sec2.flow_m, speed=x
            )
            # add 1 pascal to guarantee that discharge pressure is higher
            delta_p = point.disch.p().m - (self.guarantee_point_sec2.disch.p().m + 1)
            return delta_p

        new_speed = newton(calculate_disch_pressure_delta, self.speed_operational.m)

        return self.__class__(
            guarantee_point_sec1=self.guarantee_point_sec1,
            guarantee_point_sec2=self.guarantee_point_sec2,
            test_points_sec1=self.test_points_sec1,
            test_points_sec2=self.test_points_sec2,
            speed_operational=new_speed,
            reynolds_correction=self.reynolds_correction,
            bearing_mechanical_losses=self.bearing_mechanical_losses,
        )


@check_units
def k_seal(flow_m, state_up, state_down):
    """Function to calculate the seal constant k.

    This seal constant is used to estimate the seal leakage in different conditions.

    Parameters
    ----------
    flow_m : float, pint.Quantity
        Mass flow through the seal (kg/s).
    state_up : ccp.State
        State upstream of the seal.
    state_down : ccp.State
        State downstream of the seal.

    Returns
    -------
    k_end_seal : pint.Quantity
        Seal constant (kelvin ** 0.5 * kilogram ** 0.5 * mole ** 0.5 / pascal / second).
    """
    p_up = state_up.p()
    z_up = state_up.z()
    T_up = state_up.T().to("K")
    MW = state_up.molar_mass()
    p_down = state_down.p()

    k = (flow_m * np.sqrt(z_up * T_up / MW)) / (
        p_up * np.sqrt(1 - (p_down / p_up) ** 2)
    )

    return k


@check_units
def flow_m_seal(k_seal, state_up, state_down):
    """Function to calculate the mass flow across a seal with constant k_seal.

    Parameters
    ----------
    k_seal : pint.Quantity
        Seal constant (kelvin ** 0.5 * kilogram ** 0.5 * mole ** 0.5 / pascal / second).
    state_up : ccp.State
        State upstream of the seal.
    state_down : ccp.State
        State downstream of the seal.

    Returns
    -------
    flow_m_seal : pint.Quantity
        Seal mass flow (kg/s).
    """
    p_up = state_up.p()
    z_up = state_up.z()
    T_up = state_up.T().to("K")
    MW = state_up.molar_mass()
    p_down = state_down.p()

    flow_m_seal = (
        k_seal
        * p_up
        * (np.sqrt(1 - (p_down / p_up) ** 2))
        / (np.sqrt(z_up * T_up / MW))
    )

    return flow_m_seal


def parameter_interpolation(phi, phi_values, parameter_values):
    """Function used to make a linear interpolation for Mach and Reynolds numbers.

    Parameters
    ----------
    phi : float
        Value of phi to interpolate.
    phi_values : list
        List of phi values.
    parameter_values : list
        List of parameter values (Mach or Reynolds).
    """

    if phi < min(phi_values):
        phi_0 = phi_values[0]
        phi_1 = phi_values[1]
    elif phi > max(phi_values):
        phi_0 = phi_values[-2]
        phi_1 = phi_values[-1]
    else:
        phi_0 = [phi_ for phi_ in phi_values if (phi_ - phi) < 0][-1]
        phi_1 = [phi_ for phi_ in phi_values if (phi_ - phi) > 0][0]

    # get phi_values index
    idx0 = phi_values.index(phi_0)
    idx1 = phi_values.index(phi_1)
    parameter0 = parameter_values[idx0]
    parameter1 = parameter_values[idx1]

    # do linear interpolation for phi and parameter
    result = parameter0 + (parameter1 - parameter0) * (phi - phi_0) / (phi_1 - phi_0)

    return result



import numpy as np
import pandas as pd
import pytest
from pathlib import Path
from numpy.testing import assert_allclose
from pandas.testing import assert_frame_equal
import ccp
import ccp.data_io

Q_ = ccp.Q_

test_dir = Path(__file__).parent
data_dir = test_dir / "data"


@pytest.fixture
def impeller_lp_sec1_from_csv():
    suc = ccp.State(
        p=Q_(4.08, "bar"),
        T=Q_(33.6, "degC"),
        fluid={
            "METHANE": 58.976,
            "ETHANE": 3.099,
            "PROPANE": 0.6,
            "N-BUTANE": 0.08,
            "I-BUTANE": 0.05,
            "N-PENTANE": 0.01,
            "I-PENTANE": 0.01,
            "NITROGEN": 0.55,
            "HYDROGEN SULFIDE": 0.02,
            "CARBON DIOXIDE": 36.605,
        },
    )
    imp = ccp.Impeller.load_from_engauge_csv(
        suc=suc,
        curve_name="lp-sec1-caso-a",
        curve_path=data_dir,
        b=Q_(5.7, "mm"),
        D=Q_(550, "mm"),
        head_units="kJ/kg",
        flow_units="m³/h",
        speed_units="RPM",
        number_of_points=7,
    )

    return imp


def test_impeller_lp_sec1_from_csv(impeller_lp_sec1_from_csv):
    p0 = impeller_lp_sec1_from_csv[0]
    assert_allclose(p0.flow_v.to("m³/h").m, 11250)
    assert_allclose(p0.speed.to("RPM").m, 6882.0)
    assert_allclose(p0.disch.p().to("bar").m, 9.028963080391938)
    assert_allclose(p0.head.to("kJ/kg").m, 82.87008516878088)
    assert_allclose(p0.eff.m, 0.789412, rtol=1e-5)
    assert_allclose(p0.power.to("kW").m, 1432.5554941076675, rtol=1e-4)


def test_fluctuation():
    assert_allclose(ccp.data_io.fluctuation(np.array([1, 2, 3, 4, 5])), 133.333333)
    assert_allclose(ccp.data_io.fluctuation(np.array([1, 1, 1, 1, 1])), 0.0)


def test_fluctuation_data():
    df = pd.DataFrame({"a": [1, 2, 3, 4], "b": [4, 5, 6, 7]})
    assert_frame_equal(
        ccp.data_io.fluctuation_data(df),
        pd.DataFrame(
            {"a": [100.0, 66.66666666666667], "b": [40.0, 33.333333333333336]},
            index=[2, 3],
        ),
    )


def test_mean_data():
    df = pd.DataFrame({"a": [1, 2, 3, 4], "b": [4, 5, 6, 7]})
    assert_frame_equal(
        ccp.data_io.mean_data(df),
        pd.DataFrame({"a": [2.0, 3.0], "b": [5.0, 6.0]}, index=[2, 3]),
    )


def test_filter_data():
    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 4.01, 4.02, 5, 6.01, 6.02, 6.04, 6.05],
            "b": [4, 5, 6, 7, 7.01, 7.02, 8, 9, 10, 11, 12],
        }
    )
    assert_frame_equal(
        ccp.data_io.filter_data(df, data_type={"a": "pressure", "b": "temperature"}),
        pd.DataFrame(
            index=pd.Index([5]), data={"a": [4.01], "b": [7.01], "valid": [True]}
        ),
    )


def test_filter_data_flag():
    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 4.01, 4.02, 5, 6.01, 6.02, 6.04, 6.05],
            "b": [4, 5, 6, 7, 7.01, 7.02, 8, 9, 10, 11, 12],
        }
    )
    assert_frame_equal(
        ccp.data_io.filter_data(
            df,
            data_type={"a": "pressure", "b": "temperature"},
            drop_invalid_values=False,
        ),
        pd.DataFrame(
            index=pd.Index(range(2, 11, 1)),
            data={
                "a": [
                    2.0,
                    3.0,
                    3.67,
                    4.01,
                    4.343333333333333,
                    5.01,
                    5.676666666666667,
                    6.023333333333333,
                    6.036666666666666,
                ],
                "b": [
                    5.0,
                    6.0,
                    6.669999999999999,
                    7.010000000000001,
                    7.343333333333334,
                    8.006666666666666,
                    9.0,
                    10.0,
                    11.0,
                ],
                "valid": [
                    False,
                    False,
                    False,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
            },
        ),
    )


def test_filter_real_data():
    data_path = Path(ccp.__file__).parent / "tests/data"
    # load data.parquet
    df = pd.read_parquet(data_path / "data.parquet")
    data_type = {
        "ps": "pressure",
        "Ts": "temperature",
        "pd": "pressure",
        "Td": "temperature",
        "speed": "speed",
        "delta_p": "delta_p",
    }

    df = ccp.data_io.filter_data(
        df,
        data_type=data_type,
    )

    assert len(df) == 2



import numpy as np
import pandas as pd
import pytest
from pathlib import Path
from numpy.testing import assert_allclose
import ccp
from ccp import Q_

test_dir = Path(__file__).parent
data_dir = test_dir / "data"


@pytest.fixture
def suc_0():
    patm = Q_(101.3, "kPa")

    suc = ccp.State(
        p=Q_(1599, "kPa") + patm,
        T=Q_(40, "degC"),
        fluid={"NITROGEN": 0.96, "OXYGEN": 0.04},
    )
    return suc


@pytest.fixture
def imp_0(suc_0):
    imp = ccp.Impeller.load_from_engauge_csv(
        suc=suc_0,
        curve_name="p78-main-N2",
        curve_path=data_dir,
        b=Q_(4, "mm"),
        D=Q_(375.13, "mm"),
        head_units="kJ/kg",
        power_shaft_units="kW",
        power_losses_units="kW",
        flow_units="m³/min",
        number_of_points=8,
    )
    return imp


@pytest.fixture
def test_point_0(imp_0):
    return imp_0.point(Q_(214.9, "m**3/min"), speed=Q_(13311, "rpm"))


def test_load_power_shaft(suc_0, test_point_0):
    # desired values: vendor results
    patm = Q_(101.3, "kPa")
    assert_allclose(test_point_0.disch.p().to("kPa").m - patm.m, 4770, atol=14)
    assert_allclose(test_point_0.disch.T().to("degC").m, 188.2, atol=1.3)

    assert_allclose(test_point_0.power.to("kW").m, 10242, atol=41)
    assert_allclose(test_point_0.power_shaft.to("kW").m, 10345, atol=43)
    assert_allclose(test_point_0.head.to("kJ/kg").m, 120.1, atol=1)
    assert_allclose(test_point_0.eff.m, 0.7721, atol=1e-2)
    assert test_point_0.speed.to("rpm").m == 13311
    assert_allclose(test_point_0.torque.to("N*m").m, 7421, atol=31)
    assert_allclose(suc_0.molar_mass().to("kg/kmol").m, 28.17, atol=1e-2)
    assert_allclose(
        test_point_0.flow_v.to("m**3/min").m * suc_0.rho().to("kg/m**3").m,
        3950,
        atol=8.4,
    )



"""Data processing functions for ccp."""

import pandas as pd


def fluctuation(x):
    """Calculate fluctuation of x.

    Fluctuation as defined in the ASME PTC 10-1997 is the highest reading minus
    the lowest reading divided by the mean of the readings.

    Parameters
    ----------
    x : array-like
        Array of values.

    Returns
    -------
    float
        Fluctuation of x.

    Examples
    --------
    >>> fluctuation([1, 2, 3, 4, 5])
    80.0
    >>> fluctuation([1, 1, 1, 1, 1])
    0.0
    """
    if x.mean() == 0:
        return 100
    else:
        return 100 * (x.max() - x.min()) / x.mean()


def fluctuation_data(df, window=3):
    """Calculate fluctuation of dataframe columns.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe with data to be filtered.
    window : int, optional
        Window size for rolling calculation. The default is 3.

    Returns
    -------
    pandas.DataFrame
        Dataframe with fluctuation values.


    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    >>> fluctuation_data(df)
         a    b
    0  0.0  0.0
    1  0.0  0.0
    2  0.0  0.0
    >>> df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    >>> fluctuation_data(df, window=2)
            a    b
    0     0.0  0.0
    1  1000.0  0.0
    2  1000.0  0.0
    """
    fluctuation_df = (
        df.apply(pd.to_numeric)
        .rolling(
            window=window,
        )
        .apply(fluctuation)
        .fillna(0.0)
    )
    fluctuation_df = fluctuation_df[window - 1 :]
    return fluctuation_df


def mean_data(df, window=3):
    """Calculate the mean of dataframe columns.

    The mean is calculated using a rolling window.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe with data to be filtered.
    window : int, optional
        Window size for rolling calculation. The default is 3.

    Returns
    -------
    pandas.DataFrame
        Dataframe with mean values.


    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    >>> fluctuation_data(df)
         a    b
    0  0.0  0.0
    1  0.0  0.0
    2  0.0  0.0
    >>> df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    >>> fluctuation_data(df, window=2)
            a    b
    0     0.0  0.0
    1  1000.0  0.0
    2  1000.0  0.0
    """
    mean_df = (
        df.apply(pd.to_numeric)
        .rolling(
            window=window,
        )
        .mean()
    )
    mean_df = mean_df[window - 1 :]
    return mean_df


def filter_data(
    df,
    window=3,
    data_type=None,
    temperature_fluctuation=0.5,
    pressure_fluctuation=2,
    speed_fluctuation=0.5,
    drop_invalid_values=True,
):
    """Filter data according to fluctuation values.

    This function performs two filters:
        1. Remove rows where the speed is zero.
        2. Remove rows where the fluctuation of the data is higher than the
        maximum fluctuation defined by the user.

    After filtering, it returns the mean value for set of values defined by
    the window size.
    Default values for maximum fluctuation are based on ASME PTC 10-1997.
    As per ASME PTC 10-1997, the minimum duration of a test point is 15 minutes.
    Assuming that we need a minimum of 3 measurements to calculate the fluctuation, the
    time span between each measurement is 7.5 minutes.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe with data to be filtered.
    window : int, optional
        Window size for rolling calculation, meaning how many rolls will be used
        to calculate the fluctuation.
        The default is 3.
    data_type : dict
        Dictionary with data types for each column.
        Values for data_type can be: "pressure", "temperature", "speed".
    temperature_fluctuation : float, optional
        Maximum fluctuation for temperature data.
        The default is 0.5.
    pressure_fluctuation : float, optional
        Maximum fluctuation for pressure data.
        The default is 2.
    speed_fluctuation : float, optional
        Maximum fluctuation for speed data.
        The default is 0.5.
    drop_invalid_values : bool, optional
        Drop invalid values from the dataframe.
        If false, a column 'valid' will be added to the dataframe with True for valid.
        The default is True.

    Returns
    -------
    pandas.DataFrame
        Filtered dataframe.

    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({'a': [1, 2, 3, 4, 4.01, 4.02], 'b': [4, 5, 6, 6.01, 6.02, 6.03]})
    >>> data_type = {'a': 'pressure', 'b': 'temperature'}
    >>> filter_data(df, window=3, data_type=data_type)
    """
    fluctuation_df = fluctuation_data(df, window=window)
    mean_df = mean_data(df, window=window)
    mean_df["valid"] = True
    # filter mean_df based on fluctuation_df max values
    if drop_invalid_values:
        for column, property_type in data_type.items():
            if column in mean_df.columns:
                if property_type == "pressure":
                    max_fluctuation = pressure_fluctuation
                    # remove pressure values below 0
                    mean_df.loc[mean_df[column] < 0, column] = None
                elif property_type == "temperature":
                    max_fluctuation = temperature_fluctuation
                    # remove temperature values below 0
                    mean_df.loc[mean_df[column] < 0, column] = None
                elif property_type == "speed":
                    max_fluctuation = speed_fluctuation
                    # remove speed values below 1
                    mean_df.loc[mean_df[column] < 1, column] = None
                elif property_type == "delta_p":
                    max_fluctuation = 100
                    # remove pressure values equal to 0
                    if "delta_p" in mean_df.columns:
                        mean_df.loc[mean_df[column] <= 0, column] = None
                else:
                    raise ValueError(
                        f"Invalid data type for column {column}. "
                        "Valid data types are: pressure, temperature and speed."
                    )
                mean_df.loc[fluctuation_df[column] > max_fluctuation, column] = None
        mean_df = mean_df.dropna()
    else:
        for column, property_type in data_type.items():
            if column in mean_df.columns:
                if property_type == "pressure":
                    max_fluctuation = pressure_fluctuation
                    # remove pressure values below 0
                    mean_df.loc[mean_df[column] < 0, "valid"] = False
                elif property_type == "temperature":
                    max_fluctuation = temperature_fluctuation
                    # remove temperature values below 0
                    mean_df.loc[mean_df[column] < 0, "valid"] = False
                elif property_type == "speed":
                    max_fluctuation = speed_fluctuation
                    # remove speed values below 1
                    mean_df.loc[mean_df[column] < 1, "valid"] = False
                elif property_type == "delta_p":
                    max_fluctuation = 100
                    # remove pressure values equal to 0
                    if "delta_p" in mean_df.columns:
                        mean_df.loc[mean_df[column] <= 0, "valid"] = False
                else:
                    raise ValueError(
                        f"Invalid data type for column {column}. "
                        "Valid data types are: pressure, temperature and speed."
                    )
                mean_df.loc[fluctuation_df[column] > max_fluctuation, "valid"] = False
    return mean_df
